!
! Module: exposure_aux
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This module contains routines used in exposure calculations.
! 
! Comments:
!	TODO:***** write in the headers of all appropriate subroutines/functions (i) what cal variables they need to have set; (ii) for the calling program to deallocate any returned pointers. **** ALSO!! when it is necessary to call CAL_getState, the previous state should be stored and reloaded at the end of the routine.
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module exposure_aux

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the '*' from the lib*.a name
! in <package>/src/Makefile.

  use cal		! contains 'use caloaldefs' and 'use caltypes'
  use dal
  use errorHandling
!  use dss
!  use eventAttributes	! in saslib/sasdal. For checking which flags have been applied.

  use array_utils	! for smallestTrueIndex, largestTrueIndex,&
    ! trapezoidSum, getGridDiffs, interpolateGridDiff.
!  use constants
  use dal_aux
!  use dss_aux
!  use dss_ranges_aux
!  use geometric_types
  use intervals_aux	! for isWithinInterval() & andIntervals().
!  use splines
  use ssc_misc
!  use wcs_aux
!use dump

  implicit none 

  private 

!  public constrainSpectrum
  public correctForGtiAndAlias

!  type, public :: frameType
  type :: frameType
    real(single)    :: fracExp, deadTimeFraction
    real(double)    :: time
    type(IntervalT) :: interval
    logical(bool)   :: mask
  end type frameType

!type :: bufferedFrameType
!    type(frameType), pointer :: frames(:)
!end type bufferedFrameType

!  real(double), parameter, public ::&
!    minEeV = 50.0,&
!    maxEeV = 20000.0,&
!    binSizeInLnEeV = 0.3,&
!    coarseVigThetaArcsecBinSize = 60.0,&
!    coarseVigPhiBinSize = Pi / 18.0

contains 

!  !---------------------------------------------------------------------
!  subroutine constrainSpectrum(inBlock, spectrum, dssConstraints)
!  !---------------------------------------------------------------------

! This has been put in temp_code until I come to debug it.

!  end subroutine constrainSpectrum

  !---------------------------------------------------------------------
  subroutine correctForGtiAndAlias(exposuSet, exposuTabName, tsName&
  , gtis, treatGtis, treatCosmicRays, treatDeadTime, treatAlias&
  , frameDuration, fracExp, antiAlias, badBins, binPrecisionArg)
  !---------------------------------------------------------------------
  ! This subroutine calculates the fractional exposure for each time series
  ! bin due to the following causes:
  !	- aliasing, if event times have not been randomized within each
  ! frame;
  !	- gtis;
  !	- cosmic rays;
  !	- frame shift time.
  ! The last three are optional, depending on the value of the appropriate
  ! flag of desiresVector. The amount of aliasing is returned in antiAlias
  ! and the multiple of the other three exposure factors is returned in
  ! fracExp.
  !
  ! How these things are calculated depends on whether the event times
  ! have been randomized within the frames or not. To illustrate this,
  ! consider the following hypothetical example:
  !
  !	                   Frames:
  !
  !	Time (s): 0          11         22         33        43
  !	          |....x....||....x....||....x....||....x....|
  !
  !	                   Bin:
  !
  !	    Time (s): 4                                  39
  !	              |..................................|
  !
  !	                   GTI:
  !
  !	 Time (s):                        24
  !	          <.......................|
  !
  ! It can be seen that each frame is 10 seconds long, with a gap of
  ! 1 second shift time between frames. The events are imagined to arrive
  ! at a steady rate of 1 per second, with 0.2 cosmic rays per second
  ! (that is, the FRACEXP values in the EXPOSUnn table are all equal
  ! to 0.8). Hence 8 events are detected per frame.
  !
  ! Non-randomized case:
  ! --------------------
  ! In this case each event time is recorded as the midpoint of the frame
  ! in which it occurs (the points designated by an x in the diagram).
  ! After filtering with the GTI the un-exposure-corrected rate r_raw in
  ! the bin is N/t_bin, where N is the number of events (here 16, since
  ! only the first 2 frames are passed by the GTI) and t_bin is the
  ! duration of the time series bin (here 35 s). Thus r_raw = 16/35
  ! = 0.45714.
  !
  ! The effect of the GTI in the present case is to reduce the number of
  ! frames whose centers fall within the bin, and therefore the number of
  ! events, by a factor of 2. Hence the contribution of the GTI to fracexp
  ! is 0.5.
  !
  ! The contribution due to cosmic rays is just formed as the average over
  ! the FRACEXP values of the two frames whose centers fall within the
  ! bin & GTI. This gives 0.8.
  !
  ! The contribution of the 'dead' frame shift time is clearly going to be
  ! equal to 10/11 = 0.90909, since each frame whose center falls within
  ! bin & GTI would contribute 1 more event if it were not for the shift
  ! time. This should be calculated as
  !
  !		 1              [    TIMEDEL(i)       ]
  !		--- * sum(i=1,2)[---------------------].
  !		 2              [ TIME(i+1) - TIME(i) ]
  !
  ! The aliasing is equal to the sum of frame time 'allocated' to the whole
  ! bin (that is, the sum of TIME(i+1) - TIME(i) for all frames whose
  ! centers fall within the bin) divided by the bin duration. This equals
  ! 44/35 = 1.25714. Note that, in contrast to the other factors, it may
  ! be larger than zero.
  !
  ! The 'natural' rate of 1 event/sec ought to be returned if we divide
  ! r_raw by each of the three factors; in fact we see that this is true,
  ! since
  !
  !	0.45714 / 0.5 / 0.8 / 0.90909 / 1.25714 = 1.0.
  !
  ! Randomized case:
  ! ----------------
  ! Here the event times are dithered throughout the duration (TIMEDEL) of
  ! the frame they occur in. In the limit of high event rate (or, equiv-
  ! alently, taking an ensemble average) the events may be thought of as
  ! uniformly distributed through a period of TIMEDEL/2 either side of
  ! the TIME value. After filtering with the GTI there are (on ensemble
  ! average) 14.4 events that fall into the bin, giving a value of r_raw
  ! for that bin of 14.4/35 = 0.41143.
  !
  ! GTI contribution: this is calculated by dividing the actual amount of
  ! frame time that fell within the bin by the amount that _would_ have
  ! fallen in the bin without the gti. Here this is 18/32 = 0.5625.
  !
  ! The cosmic ray contribution should be formed from a suitably weighted
  ! average of the frame FRACEXP values (in fact the present algorithm is
  ! just to average FRACEXP for those frames which overlap the bin) but in
  ! this simplistic example it remains at 0.8.
  !
  ! The dead time contribution should be calculated as (bin duration -
  ! - total gaps between frames) / (bin duration), = 32/35 = 0.91428 in
  ! this example.
  !
  ! (Note that there is another way to 'cut the cake', depending on which
  ! correction is notionally applied first. Ie we could define the dead
  ! time exposure as 18/20 = 0.9 and the gti exposure as 20/35 = 0.57143.
  ! This is essentially a conceptual issue and doesn't affect the outcome.)
  !
  ! Since the event times are randomized, there is no aliasing and the
  ! contribution from this is therefore set at 1.0.
  !
  ! Again it can be seen that the natural rate divided by all four factors
  ! comes out as expected, since
  !
  !	0.41143 / 0.5625 / 0.8 / 0.91428 / 1.0 = 1.0.
  !
  !---------------------------------------------------------------------

    implicit none 

    type(DataSetT),  intent(in)           :: exposuSet
    character(*),    intent(in)           :: exposuTabName,&
                                             tsName
!exposuSetName,
    type(IntervalT), intent(in)           :: gtis(:)
    logical(bool),   intent(in)           :: treatGtis,&
                                             treatCosmicRays,&
                                             treatDeadTime,&
                                             treatAlias
    real(double),    intent(in)           :: frameDuration
    real(single),    intent(out)          :: fracExp(:),&
                                             antiAlias(size(fracExp))
    logical(bool),   intent(inout)        :: badBins(size(fracExp))
    real(double),    intent(in), optional :: binPrecisionArg

    type(DataSetT) :: tsSet
!exposuSet, 
    type(TableT) :: exposuTab, tsTab
    logical :: eventTimesRandomized, localTreatGtis
    integer(int32) ::&
      numBins, numGtis, numFrames, bin, startBin, finisBin,&
      startFrame, stopFrame, frameNum,&
      numFramesInBinAndGti, numFramesInBuffer, firstFrameInBuffer,&
      lastFrameInBuffer, gi, instrumentId, badFrameI
    real(single), pointer :: timeDel(:)
!    real(single), allocatable :: frameWidth(:)
    real(double), pointer :: time(:)
    real(double), allocatable :: exposuTimeDel(:), binEdges(:)
    type(IntervalT) :: binterval
    type(IntervalT), pointer ::&
      binFrames(:), binGtis(:), gtiFrames(:), miniFrames(:)
    type(frameType), allocatable :: frames(:)
    real(double) ::&
      timeInBin, liveTimeInBin, timeInBinGtis, liveTimeInBinGtis, liveTime,&
      binPrecision
    real(single) ::&
      numerator, denominator, gtiExposure, cosmicRayExposure, deadTimeExposure,&
      minBinDuration, binPrecisionPercent
    logical(bool), allocatable :: miniFrameMask(:), gtiMask(:)
    integer, parameter :: numCertainDigits = 11
!integer :: i
    character(20) :: instrumentName

logical :: testFlag = .false.
logical :: doDummy = .false.

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, '  Entering subroutine &
      &correctForGtiAndAlias:')

if (doDummy) then
  call warning('correctForGtiAndAlias', "This is a dummy routine! It &
    &doesn't return the real fracexp and antiAlias.")
  fracExp = 1.0
  antiAlias = 1.0
return
end if

if (testFlag) print *, '  start correctForGtiAndAlias.'

    !...................................................................
    ! Load timeseries bin time data:

    call message(AppMsg, Verbose, '    Opening the time series.')

    tsSet = dataSet(tsName, READ)
    tsTab = table(tsSet, 'RATE')
    numBins = numberOfRows(tsTab)
    if (numBins .ne. size(fracExp))&
call error('badSizeFracExpVector',"The vector &
&argument fracexp has a size unequal to the number of rows of the RATE table.")

    allocate(time(0:numBins-1), timeDel(0:numBins-1))
    call readColDataToFixed(tsTab, 'TIME', time)
    ! in ssclib/dal_aux.

    if (hasColumn(tsTab, 'TIMEDEL')) then
      call readColDataToFixed(tsTab, 'TIMEDEL', timeDel)
      ! in ssclib/dal_aux.
    else
      timeDel = real64Attribute(tsTab, 'TIMEDEL')
    end if
    call release(tsSet)

if (testFlag) print *, '  opened ts TIME, TIMEDEL columns'

    allocate(binEdges(0:numBins))
    binEdges(0)       = time(0)         - timeDel(0)         / 2.0d0
    binEdges(1:numBins-1) = (time(1:numBins-1) + time(0:numBins-2)) / 2.0d0
    binEdges(numBins) = time(numBins-1) + timeDel(numBins-1) / 2.0d0

    if (treatAlias) then
      if (present(binPrecisionArg)) then
        binPrecision = binPrecisionArg
      else
        binPrecision = (10.0d0**(-numCertainDigits))&
          * time(numBins-1) * 2.5d0
      end if

      minBinDuration = minval(timeDel)
      if (binPrecision > minBinDuration / 2.0)&
call error('binPrecisionTooLarge',"The bin &
&precision should be at least twice the minimum value of TIMEDEL for proper &
&correction.")

      binPrecisionPercent = 200.0 * binPrecision / minBinDuration

      write(errStr, '(a, es10.4, a, f5.2, a)') '    Bin time precision = '&
        , binPrecision, ' s (', binPrecisionPercent, '% of bin duration)'
      call message(AppMsg, Verbose)
    end if

    !...................................................................
    ! Load gtis:

    localTreatGtis = treatGtis
    if (treatGtis) then
      numGtis = size(gtis)
      if (numGtis <= 0) localTreatGtis = .false.
      allocate(gtiMask(numGtis))
    end if
if (testFlag) print *, '  loaded gtis.'

    call message(AppMsg, Verbose, '    Opening the EXPOSU extension.')

    exposuTab = table(exposuSet, exposuTabName)
    call message(AppMsg, Verbose, '    Reading RAND_TIM keyword.')
    eventTimesRandomized = .false. ! default
    if (hasAttribute(exposuTab, 'RAND_TIM')) then
      if (booleanAttribute(exposuTab, 'RAND_TIM')) then
        eventTimesRandomized = .true.
      end if
    end if
if (testFlag) print *, '  checked RAND_TIM: eventTimesRandomized =', eventTimesRandomized

    instrumentName = stringAttribute(exposuTab, "INSTRUME")
    instrumentId = instId(instrumentName) ! in caloalutils/src/caloaldefs.f90

    if (.not. eventTimesRandomized .and. instrumentId == EPN) then
      call warning('shouldRandomizeTimes',"Because PN event lists may have &
        &gaps in the frame sequence, it is not possible to ensure accurate &
        &exposure correction of PN time series, unless the event arrival &
        &times are dithered within the frame duration. This does not seem &
        &to have been done for the present event list. In order to avoid &
        &the gross errors that frame-sequence gaps can impose, lccorr will &
        &treat the event list as if the times _had_ been dithered. The &
        &difference between the dithered and non-dithered exposure treatments &
        &become significant only if the time series bin duration is small &
        &enough to be of the same order as the frame duration. If this is &
        &the case, you are advised to dither the event times, using for &
        &example the sas task etimedither, then remake the time series etc.")
    end if

    !...................................................................
    ! Load frame data:

    call message(AppMsg, Verbose, '    Reading the frame data.')

    numFrames = numberOfRows(exposuTab)
    if (numFrames <= 0)&
call error('tooFewFrames',"There are no rows in &
&the "//trim(exposuTabName)//" table.")
    allocate(frames(numFrames), exposuTimeDel(numFrames))
    call readColDataToFixed(exposuTab, 'TIME',    frames%time)
    ! in ssclib/dal_aux.
    if (any(frames(:numFrames-1)%time >= frames(2:)%time)) then
      badFrameI = smallestTrueIndex(frames(:numFrames-1)%time >= frames(2:)%time)
      ! in ssclib/array_utils.
      write(errstr, '(a, i7.7, a, d18.12, a, d18.12, a)') "The time of frame "&
        , badFrameI, " (", frames(badFrameI)%time&
        , ") is >= the following frame time (", frames(badFrameI+1)%time, ")"
!      write(errStr, '(a)') "The time of frame "
call error('badFrameOrder')
    end if

    if (hasAttribute(exposuTab, 'TIMEDEL')) then
      exposuTimeDel = real64Attribute(exposuTab, 'TIMEDEL')
    else
      call readColDataToFixed(exposuTab, 'TIMEDEL', exposuTimeDel)
      ! in ssclib/dal_aux.
    end if
    call readColDataToFixed(exposuTab, 'FRACEXP', frames%fracExp)
    ! in ssclib/dal_aux.
if (testFlag) then
  print *, '+++frames(2380)%fracExp=', frames(2380)%fracExp
  print *, '+++frames(2381)%fracExp=', frames(2381)%fracExp
  print *, '+++frames(2382)%fracExp=', frames(2382)%fracExp
end if

    call message(AppMsg, Verbose, '    Releasing the EXPOSU table.')
    call release(exposuTab)
    frames%interval%lower = frames%time - exposuTimeDel / 2.0d0
    frames%interval%upper  = frames%time + exposuTimeDel / 2.0d0

if (testFlag) then
  print *, '  calc frameWidth.'
  print *, '  numFrames=', numFrames
end if

    frames%deadTimeFraction = exposuTimeDel / frameDuration
if (testFlag) print *, '  loaded frame data.'

    numFramesInBuffer = ceiling(2.1 * maxval(timeDel) / minval(exposuTimeDel))
if (testFlag) then
  print *, ''
  print *, '  maxval(timeDel)=', maxval(timeDel)
  print *, '  minval(exposuTimeDel)=', minval(exposuTimeDel)
  print *, '  numFramesInBuffer=', numFramesInBuffer
end if
    deallocate(exposuTimeDel)

    !...................................................................
    ! Now calculate the corrections:

    call message(AppMsg, Verbose, '    Main loop.')

    !...................................................................
    if (eventTimesRandomized .or. instrumentId == EPN) then
if (testFlag) print *, '  start of randomized section.'

      antiAlias = 1.0

      if (.not. (localTreatGtis .or. treatCosmicRays .or. treatDeadTime)) then
        fracExp = 1.0
  return
      end if

      if (binEdges(1) > frames(1)%time) then
        startBin = 1
      else
        startBin = largestTrueIndex(binEdges(1:) <= frames(1)%time)
        ! in ssclib/array_utils.

        fracExp(  :startBin-1) = 0.0
!        antiAlias(:startBin-1) = 1.0
      end if

      if (binEdges(numBins-1) < frames(numFrames)%time) then
        finisBin = numBins
      else
        finisBin = smallestTrueIndex(binEdges(0:numBins-1)&
          >= frames(numFrames)%time)
        ! in ssclib/array_utils.

        fracExp(  finisBin+1:) = 0.0
!        antiAlias(finisBin+1:) = 1.0
      end if

      firstFrameInBuffer = 1
      lastFrameInBuffer  = numFrames

if (testFlag) then
  print *, '  startBin, finisBin:', startBin, finisBin
end if

      do bin = startBin, finisBin
if (testFlag) then
  print *, ''
  print *, '>>>>>>>>>>>>>>>>>>>>'
  print *, ''
  print *, 'bin/numBins ', bin, '/', numBins
  print *, 'firstFrameInBuffer, lastFrameInBuffer:'&
    , firstFrameInBuffer, lastFrameInBuffer
  print *, 'times:           start(bin):', binEdges(bin-1)
  print *, '                   end(bin):', binEdges(bin)
  print *, '  frame(firstFrameInBuffer):', frames(firstFrameInBuffer)%time
  print *, '   frame(lastFrameInBuffer):', frames(lastFrameInBuffer)%time
end if

        errStr = sprintEvery(bin, numBins, fraction=0.1) ! in ssclib/ssc_misc.
if (testFlag) print *, 'sprintEvery succeeded. errStr=', errStr
        if (errStr .ne. '') then
          errStr = '     '//errStr
          call message(appMsg, verbose)
        end if

        binterval%lower = time(bin - 1) - timeDel(bin - 1) / 2.0
        binterval%upper = time(bin - 1) + timeDel(bin - 1) / 2.0
        timeInBin = binEdges(bin) - binEdges(bin-1)

        ! Calculate indices of first and last frames that overlap the bin:

        frames(firstFrameInBuffer:lastFrameInBuffer)%mask&
          = (   frames(firstFrameInBuffer:lastFrameInBuffer)%interval%upper&
          >= binEdges(bin-1)&
          .and. frames(firstFrameInBuffer:lastFrameInBuffer)%interval%lower&
          <  binEdges(bin))

        if (.not. any(frames(firstFrameInBuffer:lastFrameInBuffer)%mask)) then
          fracExp(bin)   = 0.0
          lastFrameInBuffer = min(lastFrameInBuffer + numFramesInBuffer&
            , numFrames)
if (testFlag) print *, 'No frames in the bin.'
      cycle
        end if

        startFrame = firstFrameInBuffer - 1&
          + smallestTrueIndex(frames(firstFrameInBuffer:lastFrameInBuffer)%mask)
        stopFrame  = firstFrameInBuffer - 1&
          +  largestTrueIndex(frames(firstFrameInBuffer:lastFrameInBuffer)%mask)
        ! these calls are in ssclib/array_utils.

if (testFlag) then
  print *, ''
  print *, 'startFrame, stopFrame:', startFrame, stopFrame
  print *, 'times:'
  print *, '        frame(startFrame):', frames(startFrame)%time
  print *, '         frame(stopFrame):', frames(stopFrame)%time
  print *, ''
  print *, '         total bin time: ', binEdges(bin) - binEdges(bin-1)
  print *, ''
end if

        if (stopFrame >= lastFrameInBuffer&
        .and. .not. (lastFrameInBuffer == numFrames))&
call error('bug0',"Please contact the code developer.")

        ! Set up buffer for next bin:
        !
        firstFrameInBuffer = stopFrame
        lastFrameInBuffer  = min(stopFrame + numFramesInBuffer, numFrames)

        ! Now proceed with the exposure calculations:
        !
        call andIntervals(binterval, frames(startFrame:stopFrame)%interval&
          , binFrames) ! in ssclib/intervals_aux.

        liveTimeInBin = sum(binFrames%upper) - sum(binFrames%lower)
        ! This has to be > 0, from the way the frame start and stop indices
        ! were defined.

if (testFlag) then
  print *, '          liveTimeInBin:', liveTimeInBin
end if
        if (localTreatGtis) then
          call andIntervals(binterval, gtis, binGtis) ! Not going to worry
          ! about limiting the gti vector to between start and stop indices,
          ! as we did with the frames. This is because there are usually
          ! far fewer gtis than frames.
          !
          ! andIntervals is in ssclib/intervals_aux.

          if (size(binGtis) <= 0) then
            fracExp(bin)   = 0.0
            deallocate(binGtis, binFrames)
      cycle
          end if

          timeInBinGtis = sum(binGtis%upper)&
            - sum(binGtis%lower)

          call andIntervals(binGtis, frames(startFrame:stopFrame)%interval&
            , gtiFrames) ! in ssclib/intervals_aux.

          liveTimeInBinGtis = sum(gtiFrames%upper) - sum(gtiFrames%lower)
          gtiExposure = liveTimeInBinGtis / liveTimeInBin
if (testFlag) then
  print *, '          timeInBinGtis:', timeInBinGtis
  print *, '      liveTimeInBinGtis:', liveTimeInBinGtis
  print *, ''
  print *, 'binGtis(1)%lower=', binGtis(1)%lower
  print *, 'binGtis(n)%upper=', binGtis(size(binGtis))%upper
  print *, 'num binGtis=', size(binGtis)
  print *, ''
  print *, '      gtiExposure:', gtiExposure
end if
          deallocate(gtiFrames)

          if (treatCosmicRays) then ! form a weighted average of all frames
          ! which overlap the within-bin GTIs:

            numerator   = 0.0
            denominator = 0.0
            do frameNum = startFrame, stopFrame
              call andIntervals(frames(frameNum)%interval, binGtis, miniFrames)
              ! in ssclib/intervals_aux.
              liveTime = sum(miniFrames%upper) - sum(miniFrames%lower)
              numerator   = numerator   + liveTime * frames(frameNum)%fracExp
              denominator = denominator + liveTime
              deallocate(miniFrames)
            end do

            cosmicRayExposure = numerator / denominator

          else
            cosmicRayExposure = 1.0
          end if

          deallocate(binGtis)

          if (treatDeadTime) then
!            deadTimeExposure = liveTimeInBinGtis / timeInBinGtis
            deadTimeExposure = liveTimeInBin / timeInBin
          else
            deadTimeExposure = 1.0
          end if

        else ! don't treat GTIs
          gtiExposure = 1.0

          if (treatCosmicRays) then ! form a weighted average of all frames
          ! which overlap the bin:

            numerator   = 0.0
            denominator = 0.0
            do frameNum = startFrame, stopFrame
              call andIntervals(binterval, frames(frameNum)%interval&
                , miniFrames) ! in ssclib/intervals_aux.

              liveTime = sum(miniFrames%upper) - sum(miniFrames%lower)
              numerator   = numerator   + liveTime * frames(frameNum)%fracExp
              denominator = denominator + liveTime
              deallocate(miniFrames)
            end do

            cosmicRayExposure = numerator / denominator

          else
            cosmicRayExposure = 1.0
          end if

          if (treatDeadTime) then
            deadTimeExposure = liveTimeInBin / timeInBin
          else
            deadTimeExposure = 1.0
          end if
        end if ! treat GTIs

        deallocate(binFrames)

        fracExp(bin) = gtiExposure * cosmicRayExposure * deadTimeExposure
if (testFlag) then
  print *, 'cosmicRayExposure:', cosmicRayExposure
  print *, ' deadTimeExposure:', deadTimeExposure
  print *, '  => fracExp(bin):', fracExp(bin)
end if
      end do ! bin

    !...................................................................
    else ! event times were not randomized. Need to calculate anti-aliasing
    ! correction.

if (testFlag) print *, '  start of not-randomized section.'

!**** have to add something which gives 0 exposure if the frame centre time is too close to a bin edge.

      if (binEdges(1) > frames(1)%time) then
        startBin = 1
      else
        startBin = largestTrueIndex(logical(binEdges(1:)&
          <= frames(1)%time, bool))
        ! in ssclib/array_utils.

        fracExp(  :startBin-1) = 0.0
        antiAlias(:startBin-1) = 1.0
      end if

      if (binEdges(numBins-1) < frames(numFrames)%time) then
        finisBin = numBins
      else
        finisBin = smallestTrueIndex(logical(binEdges(0:numBins-1)&
          >= frames(numFrames)%time, bool))
        ! in ssclib/array_utils.

        fracExp(  finisBin+1:) = 0.0
        antiAlias(finisBin+1:) = 1.0
      end if

      firstFrameInBuffer = 1
      lastFrameInBuffer  = numFrames

if (testFlag) then
  print *, '  startBin, finisBin:', startBin, finisBin
!  print *, 'firstFrameInBuffer, lastFrameInBuffer:'&
!    , firstFrameInBuffer, lastFrameInBuffer
!  print *, 'times:      start(startBin):', binEdges(startBin-1)
!  print *, '              end(startBin):', binEdges(startBin)
!  print *, '                   frame(1):', frames(1)%time
!  print *, '  frame(firstFrameInBuffer):', frames(firstFrameInBuffer)%time
!  print *, '   frame(lastFrameInBuffer):', frames(lastFrameInBuffer)%time
end if

      BINLOOP: do bin = startBin, finisBin

if (testFlag) then
  print *, ''
  print *, '>>>>>>>>>>>>>>>>>>>>'
  print *, ''
  print *, 'bin ', bin
  print *, 'firstFrameInBuffer, lastFrameInBuffer:'&
    , firstFrameInBuffer, lastFrameInBuffer
  print *, 'times:           start(bin):', binEdges(bin-1)
  print *, '                   end(bin):', binEdges(bin)
  print *, '  frame(firstFrameInBuffer):', frames(firstFrameInBuffer)%time
  print *, '   frame(lastFrameInBuffer):', frames(lastFrameInBuffer)%time
end if

        errStr = sprintEvery(bin, numBins, fraction=0.1) ! in ssclib/ssc_misc.
        if (errStr .ne. '') then
          errStr = '     '//errStr
          call message(appMsg, verbose)
        end if

        ! Calculate indices of first and last frames for which the frame
        ! centre times fall within the bin:

        frames(firstFrameInBuffer:lastFrameInBuffer)%mask&
          = (frames(firstFrameInBuffer:lastFrameInBuffer)%time&
          >= binEdges(bin-1)&
          .and. frames(firstFrameInBuffer:lastFrameInBuffer)%time&
          <  binEdges(bin))

        if (.not. any(frames(firstFrameInBuffer:lastFrameInBuffer)%mask)) then
          fracExp(bin)   = 0.0

          antiAlias(bin) = 1.0 ! antialias is effected via multiplication
          ! by fracExp. The nil-exposure condition for this bin is already
          ! flagged by setting fracExp = 0, setting antiAlias = 0 also leads
          ! only to needless complication.

          lastFrameInBuffer = min(lastFrameInBuffer + numFramesInBuffer&
            , numFrames)
if (testFlag) print *, 'no frames!'
      cycle BINLOOP
        end if

        startFrame = firstFrameInBuffer - 1&
          + smallestTrueIndex(frames(firstFrameInBuffer:lastFrameInBuffer)%mask)
        stopFrame  = firstFrameInBuffer - 1&
          +  largestTrueIndex(frames(firstFrameInBuffer:lastFrameInBuffer)%mask)
        ! these calls are in ssclib/array_utils.

if (testFlag) then
  print *, ''
  print *, 'startFrame, stopFrame:', startFrame, stopFrame
  print *, 'times:'
  print *, '        frame(startFrame):', frames(startFrame)%time
  print *, '         frame(stopFrame):', frames(stopFrame)%time
  print *, ''
!  print *, 'total frame time in bin: '&
!    , sum(frames(startFrame:stopFrame)%interval%upper&
!    - frames(startFrame:stopFrame)%interval%lower)
  print *, 'total frame time in bin: '&
    , frameDuration * (stopFrame + 1 - startFrame)
  print *, '         total bin time: ', binEdges(bin) - binEdges(bin-1)
  print *, ''
end if

        if (stopFrame >= lastFrameInBuffer&
        .and. .not. (lastFrameInBuffer == numFrames)) then
call error('bug1',"Please contact the code developer.")
        end if

        firstFrameInBuffer = stopFrame
        lastFrameInBuffer  = min(stopFrame + numFramesInBuffer, numFrames)

        if (treatAlias) then
          if (abs(frames(startFrame)%time-binEdges(bin-1)) < binPrecision) then
            badBins(bin) = .true.
            if (bin > 1) badBins(bin-1) = .true.
      cycle BINLOOP
          end if
          if (abs(frames(stopFrame)%time-binEdges(bin)) < binPrecision) then
            badBins(bin) = .true.
            if (bin < numBins) badBins(bin+1) = .true.
      cycle BINLOOP
          end if

          antiAlias(bin) = frameDuration&
            * real(stopFrame + 1 - startFrame, double)&
            / (binEdges(bin) - binEdges(bin-1))
        end if ! treatAlias

if (testFlag) then
!  print *, '  bin', bin
!  print *, 'startFrame, stopFrame:', startFrame, stopFrame
!  print *, 'times:'
!  print *, '        frame(startFrame):', frames(startFrame)%time
!  print *, '         frame(stopFrame):', frames(stopFrame)%time
!  print *, 'total frame time in bin: '&
!    , sum(frames(startFrame:stopFrame)%interval%upper&
!    - frames(startFrame:stopFrame)%interval%lower)
!  print *, 'total bin time: ', binEdges(bin) - binEdges(bin-1)
  print *, 'hence antiAlias(bin) = ', antiAlias(bin)
  print *, ''
end if
        ! Now proceed with the exposure calculations:

        allocate(miniFrameMask(startFrame:stopFrame))
        if (localTreatGtis) then
          if (treatAlias) then
            ! Any GTIs start or stop within the bin?
            gtiMask = ((gtis%lower >= binEdges(bin-1) - binPrecision&
                  .and. gtis%lower <  binEdges(bin)   + binPrecision)&
                  .or. (gtis%upper >= binEdges(bin-1) - binPrecision&
                  .and. gtis%upper <  binEdges(bin)   + binPrecision))

            if (any(gtiMask)) then
              do gi = 1, numGtis
              if (.not. (gtiMask(gi))) cycle
                if (any(abs(frames(startFrame:stopFrame)%time - gtis(gi)%lower)&
                  < binPrecision)&
                .or. any(abs(frames(startFrame:stopFrame)%time- gtis(gi)%upper)&
                  < binPrecision)) then

                  badBins(bin) = .true.
                  deallocate(miniFrameMask)
      cycle BINLOOP
                end if
              end do
            end if ! any(gtiMask)
          end if ! treatAlias

          miniFrameMask = .false. ! default
          do frameNum = startFrame, stopFrame
            if (isWithinInterval(frames(frameNum)%time, gtis))& ! in ssclib/intervals_aux.
              miniFrameMask(frameNum) = .true.
          end do
          numFramesInBinAndGti = count(miniFrameMask)

          gtiExposure = numFramesInBinAndGti&
            / real(1 + stopFrame - startFrame)

if (testFlag) then
  print *, ''
!  print *, 'frames in bin + also in gti = ', miniFrameMask
  print *, 'num frames in bin + also in gti: ', count(miniFrameMask)
  print *, 'hence gtiExposure = ', gtiExposure
  print *, ''
end if

        else ! don't treat GTIs
          miniFrameMask = .true.
          numFramesInBinAndGti = 1 + stopFrame - startFrame
          gtiExposure = 1.0
        end if

        if (treatCosmicRays) then
          if (numFramesInBinAndGti > 0) then
            cosmicRayExposure = sum(frames(startFrame:stopFrame)%fracExp&
              , miniFrameMask) / real(numFramesInBinAndGti)
          else
            cosmicRayExposure = 0.0
          end if

if (testFlag) then
  print *, 'cosmicRayExposure (ave value of FRACEXP for frames in gti) = '&
    , cosmicRayExposure
  print *, ''
end if
        else
          cosmicRayExposure = 1.0
        end if

        if (treatDeadTime) then
          if (numFramesInBinAndGti > 0) then
            deadTimeExposure&
              = sum(frames(startFrame:stopFrame)%deadTimeFraction&
              , miniFrameMask) / real(numFramesInBinAndGti)

          else
            deadTimeExposure = 0.0
          end if

if (testFlag) then
  print *, 'deadTimeExposure = ', deadTimeExposure
  print *, ''
  print *, ''
end if
        else
          deadTimeExposure = 1.0
        end if

        deallocate(miniFrameMask)

        fracExp(bin) = gtiExposure * cosmicRayExposure * deadTimeExposure
      end do BINLOOP ! bin
if (testFlag) print *, '  end of main loop.'
    end if ! eventTimesRandomized
if (testFlag) print *, '  out of randomized fork.'


if (testFlag) print *, '  end correctForGtiAndAlias.'

  end subroutine correctForGtiAndAlias

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module exposure_aux
