!
! Module: exposure_map_pcms
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This module contains routines used in exposure calculations.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module exposure_map_pcms

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the '*' from the lib*.a name
! in <package>/src/Makefile.

  use cal		! contains 'use caloaldefs' and 'use caltypes'
  use dal
  use errorHandling
  use dss
  use eventAttributes	! in saslib/sasdal. For checking which flags have been applied.
  use dataFilter     	! in selectlib.

  use step_functions
!  use exposure_aux

  use array_utils	! for smallestTrueIndex
  use badpixutils_mod	! (in emsaplib) for readBadOffsets.
  use boxchains
  use constants
  use coordinate	! for detUnitArcsec
!use l_coordinate	! for detUnitArcsec
!  use dss_aux
  use dss_bitmasks_aux
  use dss_ranges_aux
  use exposure
  use geometric_types
  use sort_mod
  use ssc_misc		! for fracVal
  use wcs_aux

use random_aux ! only for dumping images.
use dump

  implicit none 

  private 

  public readCcdExposureInfo
  public getEnergyBounds
  public constructModelSpectrum
  public calcEeVFunctionNorm

  public calcCcdExpMap
  public calcPnCcdExpMap
  private getPnBadPixelMap
  public calcMosCcdExpMap
  private getMosBadPixelMap
  private calcCcdSpatialMap
  private createCoarsePseudoEvlist
  private createFinePseudoEvlist
private l_tanToPixelsVector
  public calcFlaggedMask
  private calcPnFlagMap
  private calcMosFlagMap
  private readEvatts

  interface readCcdExposureInfo
    module procedure readCcdExposureInfoSet, readCcdExposureInfoName
  end interface

  integer, parameter ::&
    BRIGHTPIX_BIT      = 0,&
    ONBOARD_BADPIX_BIT = 1,&
    DEADPIX_BIT        = 2,&
    BAD_OFFSET_BIT     = 3,&
    OUTSIDE_WINDOW_BIT = 4,&
    OUTSIDE_FOV_BIT    = 5,&
    NODE_BOUNDARY_BIT  = 6
  integer, parameter ::&
    OUT_OF_FOV_I             = 1,&
    CLOSE_TO_CCD_BORDER_I    = 2,&
    CLOSE_TO_NODE_BOUNDARY_I = 3,&
    CLOSE_TO_CCD_WINDOW_I    = 4
  integer, public, parameter ::&
    TREAT_FILTER_BIT    = 0,&
    TREAT_VIGNET_BIT    = 1,&
    TREAT_QE_BIT        = 2,&
    TREAT_GTI_BIT       = 3,&
    TREAT_COSMIC_BIT    = 4,&
    TREAT_DEAD_TIME_BIT = 5,&
    TREAT_ALIAS_BIT     = 6

  logical(bool) :: evattIsSet(4)

  type, public :: SpectrumType
    real(double) :: EeV, value
  end type SpectrumType

  type, public :: expInfoType
    logical(bool)  :: exposuExists, randXY, randT
    character(8)   :: tabName
    integer(int32) :: ccdMode
    integer(int16) :: winMinRawX, winMaxRawX, winMinRawY, winMaxRawY
    real(double)   :: tShift, tDead, tFrame, temperature, intFraction
  end type expInfoType

  type :: rangeIndexType
    integer(int32) :: lowerI, upperI
  end type rangeIndexType

  type :: boxIndexType
    type(rangeIndexType) :: x, y
  end type boxIndexType

  type :: boxIndexChainType
    type(boxIndexType), pointer :: boxes(:)
    type(BoundT), pointer :: xEdges(:), yEdges(:)
    integer(int32), pointer :: dssCompNums(:)
    character(20) :: xAxisName, yAxisName
  end type boxIndexChainType

  type, public :: localTAWtype
    real(double)                 :: timeStamp
    type(SpacecraftAttitudeType) :: attitude
    type(WcsT)                   :: wcs
  end type localTAWtype

contains 

  !---------------------------------------------------------------------
  subroutine readCcdExposureInfoName(eventSetName, ccdStr, instrumentId&
  , expInfo, filtId)
  !---------------------------------------------------------------------

    implicit none

    character(*),      intent(in)  :: eventSetName
    character(2),      intent(in)  :: ccdStr(:,:)
    integer(int32),    intent(in)  :: instrumentId
    type(expInfoType), intent(out) :: expInfo(size(ccdStr,1),size(ccdStr,2))
    integer(int32),    intent(out) :: filtId

    type(DataSetT) :: eventSet

    !-------------------------------------------------------------------
    ! Start

    eventSet = dataSet(eventSetName, READ)

    call readCcdExposureInfoSet(eventSet, ccdStr, instrumentId&
      , expInfo, filtId)

    call release(eventSet)

  end subroutine readCcdExposureInfoName

  !---------------------------------------------------------------------
  subroutine readCcdExposureInfoSet(eventSet, ccdStr, instrumentId&
  , expInfo, filtId)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT),    intent(in)  :: eventSet
    character(2),      intent(in)  :: ccdStr(:,:)
    integer(int32),    intent(in)  :: instrumentId
    type(expInfoType), intent(out) :: expInfo(size(ccdStr,1),size(ccdStr,2))
    integer(int32),    intent(out) :: filtId

    integer :: numCcds, numNodes, ccd, node
    character(20) :: eventListInstrumentStr
    integer(int32) :: eventListInstrumentId, ccdmodeFromHeader 
    logical :: noExposuExtensions
    type(TableT) :: inccdtbl
    type(BlockT) :: exposuBlock
    type(ModeParameterDataType) ::& ! typedef from caltypes.
      modeParams
    real(double) :: nominalTint, nominalTframe

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, "Reading information from the event set.")

    numCcds  = size(ccdStr,1)
    numNodes = size(ccdStr,2)

    eventListInstrumentStr = stringAttribute(eventSet, 'INSTRUME')
    eventListInstrumentId = instId(trim(eventListInstrumentStr))

    if (eventListInstrumentId .ne. instrumentId) then
      write(errstr, *) "INSTRUME keyword "//trim(instId(eventListInstrumentId))&
      //" in the event list doesn't match the one supplied."
call error('evlistWrongInstrument')
    end if

    noExposuExtensions   = .true.  ! default
    expInfo%exposuExists = .false. ! array op, default
    do ccd = 1, numCcds ! numCcds, numNodes are declared in lccorr_include
      do node = 1, numNodes ! and their values set in miscSetups.
        expInfo(ccd, node)%tabName = 'EXPOSU'//ccdStr(ccd, node)
      if (.not. hasBlock(eventSet, expInfo(ccd, node)%tabName)) cycle

        exposuBlock = block(eventSet, expInfo(ccd, node)%tabName)
        call CAL_setState(exposuBlock)
        ! This CAL_getState() seems to return SUBMODE (rather than CCDMODE)
        call CAL_getState(ccdModeId=expInfo(ccd, node)%ccdMode)
	! The results of CAL_getState() not used at least in this subroutine.

        ! copied from epframes_mod.f90 (in getParamValues(); Part 'initialize output file')
        inccdtbl =    table(eventSet, name(exposuBlock) )
        ccdmodeFromHeader = int16Attribute(inccdtbl,'CCDMODE')

        if (.not.&
            ((ccdModeFromHeader == PRIME_FULL_WINDOW)&
        .or. (ccdModeFromHeader == PRIME_PARTIAL_RFS)&
        .or. (ccdModeFromHeader == PRIME_PARTIAL_W2)&
        .or. (ccdModeFromHeader == PRIME_PARTIAL_W3)&
        .or. (ccdModeFromHeader == PRIME_PARTIAL_W4)&
        .or. (ccdModeFromHeader == PRIME_PARTIAL_W5)&
        .or. (ccdModeFromHeader == PRIME_PARTIAL_W6)&
        .or. (ccdModeFromHeader == PRIME_FULL_WINDOW_EXTENDED)&
        .or. (ccdModeFromHeader == PRIME_LARGE_WINDOW)&
        .or. (ccdModeFromHeader == PRIME_SMALL_WINDOW))) then
          write(errstr, *) "Instrument mode", ccdModeFromHeader,&
          " not supported."
call error('badCcdMode')
        end if

        expInfo(ccd, node)%winMinRawX = int16attribute(exposuBlock, 'WINDOWX0')
        expInfo(ccd, node)%winMinRawY = int16attribute(exposuBlock, 'WINDOWY0')
        expInfo(ccd, node)%winMaxRawX = expInfo(ccd, node)%winMinRawX&
          + int16attribute(exposuBlock, 'WINDOWDX') - 1
        expInfo(ccd, node)%winMaxRawY = expInfo(ccd, node)%winMinRawY&
          + int16attribute(exposuBlock, 'WINDOWDY') - 1

        expInfo(ccd, node)%randXY = .false. ! default
        if (instrumentId == EPN) then
!!!          if (hasAttribute(exposuBlock, 'RAND_XY')) then
          if (hasAttribute(exposuBlock, 'RANDOM_P')) then
!!!            if (int8attribute(exposuBlock, 'RAND_XY') == 1)&
!!!              expInfo(ccd, node)%randXY = .true.
            if (booleanAttribute(exposuBlock, 'RANDOM_P'))&
              expInfo(ccd, node)%randXY = .true.
          end if
        else
          if (hasAttribute(exposuBlock, 'RAND_XY')) then
            if (int8attribute(exposuBlock, 'RAND_XY') == 1)&
              expInfo(ccd, node)%randXY = .true.
          end if
        end if

        expInfo(ccd, node)%randT = .false. ! default
        if (hasAttribute(exposuBlock, 'RAND_TIM')) then
          if (booleanAttribute(exposuBlock, 'RAND_TIM')) then
            expInfo(ccd, node)%randT = .true.
          end if
        end if

        call CAL_getState(temperature=expInfo(ccd, node)%temperature)

        call CAL_getModeParameters(modeParams)
        nominalTint = modeParams%integrationTime / 1000.0d0
        nominalTframe = modeParams%frameTime / 1000.0d0

        ! Nominal because Tint (and therefore Tframe) can vary by some
        ! amount. The actual values will be calculated on a
        ! frame-by-frame basis from the EXPOSUnn extensions of the
        ! whole-camera event list.

!if (testFlag) call dumpModeParams(modeParams)

        expInfo(ccd, node)%tFrame = nominalTframe
        expInfo(ccd, node)%tShift = modeParams%smearedFraction&
          * nominalTint
        expInfo(ccd, node)%tDead = nominalTframe&
          * (1 - modeParams%dutyCycle)
        expInfo(ccd, node)%intFraction = nominalTint / nominalTframe
        expInfo(ccd, node)%exposuExists = .true.
        noExposuExtensions = .false.
      end do ! next node
    end do ! next ccd

    filtId = filterId(stringAttribute(eventSet, 'FILTER'))

    if (noExposuExtensions) then
      write(errstr, *) "There are no EXPOSUnn extensions in &
        &the event list. Can't do exposure corrections."
call error('noExposuExtensions')
    end if

  end subroutine readCcdExposureInfoSet

  !---------------------------------------------------------------------
  function getEnergyBounds(inBlock, ccd1Temp, timeStamp, spectrum)&
  result(andedEeVRange)
  !---------------------------------------------------------------------
  ! NOTE! The cal state must be set for instrument before this routine
  ! is called.

    implicit none

    type(BlockT),       intent(in), optional :: inBlock
    real(double),       intent(in), optional :: ccd1Temp,&
                                                timeStamp
    type(SpectrumType), intent(in), optional :: spectrum(:)
    type(RangeT)                             :: andedEeVRange

    real(double) :: tsMinChan, tsMinEeV, tsMaxChan, tsMaxEeV, oldCcdTemp
    integer(int32) :: instrumentId, oldCcdChipId, oldCcdNodeId
    type(RangeT) ::&
      calEeVrange, specEeVrange
    character(10) ::&
      tsLoEeVStr,  tsHiEeVStr,&
      specLoEeVStr, specHiEeVStr, calLoEeVStr, calHiEeVStr
    integer :: rangeOverlapState
    logical :: rangeHasBeenInitialized

    !-------------------------------------------------------------------
    ! Start

    rangeHasBeenInitialized = .false.

    !...................................................................
    ! Get initial tsEeVrange from keywords in the RATE extension:

    if (present(inBlock)) then
      rangeHasBeenInitialized = .true.

      if (hasAttribute(inBlock, 'CHANMIN')&
      .and. hasAttribute(inBlock, 'CHANMAX')) then
        tsMinChan = real64Attribute(inBlock, 'CHANMIN')
        tsMaxChan = real64Attribute(inBlock, 'CHANMAX')

        select case(stringAttribute(inBlock, 'CHANTYPE'))
        case('PI')
          tsMinEeV = convertPiToEeV(tsMinChan) ! in ssclib/exposure.
          tsMaxEeV = convertPiToEeV(tsMaxChan) ! in ssclib/exposure.

        case('PHA')
if (.not.present(ccd1Temp) .or. .not.present(timeStamp)) call error('',"")
          call CAL_getState(instrument=instrumentId)

          if (instrumentId == EPN) then ! just to stop the cal giving a warning:
            call CAL_getState(ccdChipId=oldCcdChipId&
              , temperature=oldCcdTemp)

            call CAL_setState(temperature=ccd1Temp, ccdChipId=1_int32)

            tsMinEeV = convertPhaToEeV(timeStamp, tsMinChan)
            tsMaxEeV = convertPhaToEeV(timeStamp, tsMaxChan)
            ! in ssclib/exposure.

            call CAL_setState(ccdChipId=oldCcdChipId&
              , temperature=oldCcdTemp)
          else
            call CAL_getState(ccdChipId=oldCcdChipId&
              , ccdNodeId=oldCcdNodeId, temperature=oldCcdTemp)

            call CAL_setState(temperature=ccd1Temp, ccdChipId=1_int32&
              , ccdNodeId=0_int32)

            tsMinEeV = convertPhaToEeV(timeStamp, tsMinChan)
            tsMaxEeV = convertPhaToEeV(timeStamp, tsMaxChan)
            ! in ssclib/exposure.

            call CAL_setState(ccdChipId=oldCcdChipId&
              , ccdNodeId=oldCcdNodeId, temperature=oldCcdTemp)
          end if

          call warning('doubtfulChanType',"The CHANTYPE keyword was set &
            &to `PHA'. Unfortunately this results from the default &
            &behaviour of evselect at time of writing. Although it probably &
            &is of no more than minor importance, `PI' is to be preferred &
            &I think, because it is not dependent on ccd number or &
            &temperature. In order to get a single representative &
            &conversion to eV I have used the temperature from CCD 1.")

        case default
call error('badChanType',"Unrecognized value of the CHANTYPE keyword. &
&CHANTYPE must be either `PI' or `PHA'. You can set this if you use &
&evselect to make the time series: it is controlled via the &
&--energycolumn parameter of evselect.")
        end select

        andedEeVRange%lower%type = INCLUSIVE
        andedEeVRange%lower%value = tsMinEeV
        andedEeVRange%upper%type = INCLUSIVE
        andedEeVRange%upper%value = tsMaxEeV

      else ! keywords CHANMIN and CHANMAX not found
        andedEeVRange%lower%type = UNDEFINED
        andedEeVRange%lower%value = 0.0d0
        andedEeVRange%upper%type = UNDEFINED
        andedEeVRange%upper%value = 0.0d0
      end if ! hasAttributes CHANMIN and CHANMAX.
    end if ! present(inBlock)

    !...................................................................
    ! Get the energy range spanned by the DSS:

!    if (present(dssPointer)) then ! this is a user-supplied spectrum.
!      piRangeSpan = getRangesSpan(dssPointer, 'PI') ! in ssclib/dss_aux.
!      dssEeVRange = convertPiToEeV(piRangeSpan) ! in ssclib/exposure.
!
!      if (rangeHasBeenInitialized) then
!        rangeOverlapState = checkRangeOverlap(andedEeVRange, dssEeVrange)
!
!        if (  rangeOverlapState .ne. BISWITHINA&
!        .and. rangeOverlapState .ne. BEQUALSA) then
!
!          call rangeToStr(dssEeVrange,  dssLoEeVStr, dssHiEeVStr)
!          call rangeToStr(andedEeVRange, tsLoEeVStr,  tsHiEeVStr)
!
!          write(errstr, *) "The energy selection recorded in the DSS ("&
!            //trim(dssLoEeVStr)//" eV to "//trim(dssHiEeVStr)//" eV) appears &
!            &to extend beyond the energy limits given by the CHANMIN and &
!            &CHANMAX keywords of the time series ("//trim(tsLoEeVStr)&
!            //" eV to "//trim(tsHiEeVStr)//" eV). The intersection of the &
!            &two ranges will be used. It's possible in this situation that the &
!            &calculated exposure values will be inaccurate; on the &
!            &other hand, if the discrepancy is small, the effect is unlikely &
!            &to be significant."
!
!          call warning('dubiousDssEnergyRange')
!        end if
!
!        if ( rangeOverlapState == BISTOOHIGH&
!        .or. rangeOverlapState == BISTOOLOW) then
!call error('noERangeOverlap',"The energy ranges calculated so far don't &
!&overlap the energy ranges stored in the DSS.")
!        end if
!
!        call andRanges(dssEeVrange, andedEeVRange)
!      else
!        andedEeVRange = dssEeVrange
!        rangeHasBeenInitialized = .true.
!      end if
!    end if

    !...................................................................
    ! Get specEeVrange if there is a user-submitted spectrum:

    if (present(spectrum)) then ! this is a user-supplied spectrum.
if (size(spectrum) < 2) call error('',"")

      specEeVrange%lower%type  = INCLUSIVE
      specEeVrange%lower%value = spectrum(1)%EeV
      specEeVrange%upper%type  = INCLUSIVE
      specEeVrange%upper%value = spectrum(size(spectrum))%EeV

      if (rangeHasBeenInitialized) then
        rangeOverlapState = checkRangeOverlap(specEeVrange, andedEeVRange)
        if (  rangeOverlapState .ne. BISWITHINA&
        .and. rangeOverlapState .ne. BEQUALSA) then

          call rangeToStr(andedEeVRange, tsLoEeVStr, tsHiEeVStr)
          call rangeToStr(specEeVrange, specLoEeVStr, specHiEeVStr)

          write(errstr, *) "The energy limits already established for &
            &the data ("//trim(tsLoEeVStr)&
            //" eV to "//trim(tsHiEeVStr)//" eV) appear to extend beyond &
            &the energy limits of the user-supplied spectrum ("&
            //trim(specLoEeVStr)//" eV to "//trim(specHiEeVStr)&
            //" eV). The intersection of the two ranges will be used. It's &
            &possible in this situation that the calculated exposure &
            &values will be inaccurate."

          call warning('dubiousSpecEnergyRange')
        end if

        if ( rangeOverlapState .eq. BISTOOHIGH&
        .or. rangeOverlapState .eq. BISTOOLOW) then
call error('noERangeOverlap',"The energy ranges calculated so far don't &
&overlap the energy ranges of the spectrum.")
        end if

        call andRanges(specEeVrange, andedEeVRange)

      else ! rangeHasBeenInitialized = .false.
        andedEeVRange = specEeVrange
        rangeHasBeenInitialized = .true.
      end if
    end if

    !...................................................................
    ! Get absolute E bounds from the cal:
    !
    ! NOTE! The types of the calEeVrange are NOT PERMITTED
    ! to == UNDEFINED.

!*** absolute bounds in cal - call CAL_getBinnedPIeBounds, or even perhaps CAL_psfValidityRanges??? Try a few things, see if different cal functions have different bounds. Hard-wire for the time being:

calEeVrange%lower%type = INCLUSIVE
calEeVrange%lower%value = 100.0
calEeVrange%upper%type = INCLUSIVE
calEeVrange%upper%value = 12000.0

    if (rangeHasBeenInitialized) then
      rangeOverlapState = checkRangeOverlap(calEeVrange, andedEeVRange)
      if (  rangeOverlapState .ne. BISWITHINA&
      .and. rangeOverlapState .ne. BEQUALSA) then

        call rangeToStr(andedEeVRange, tsLoEeVStr, tsHiEeVStr)
        call rangeToStr(calEeVrange, calLoEeVStr, calHiEeVStr)

        write(errstr, *) "The energy limits already established for &
          &the data ("//trim(tsLoEeVStr)&
          //" eV to "//trim(tsHiEeVStr)//" eV) appear to extend beyond &
          &the maximum extent of the energy limits defined in the cal ("&
          //trim(calLoEeVStr)//" eV to "//trim(calHiEeVStr)&
          //" eV). The intersection of the two ranges will be used, &
          &to avoid querying the cal with out-of-range energy values. It's &
          &possible in this situation that the calculated exposure &
          &values will be inaccurate."

!*** note in the doco that ts should always be made with e limits at least within the cal limits.

        call warning('dubiousCalEnergyRange')
      end if

      if ( rangeOverlapState .eq. BISTOOHIGH&
      .or. rangeOverlapState .eq. BISTOOLOW) then
call error('noERangeOverlap',"The energy ranges calculated so far don't &
&overlap the limits defined in the cal.")
      end if

      call andRanges(calEeVrange, andedEeVRange)

    else ! rangeHasBeenInitialized = .false.
      andedEeVRange = calEeVrange
!      rangeHasBeenInitialized = .true.
    end if

  end function getEnergyBounds

  !---------------------------------------------------------------------
  subroutine constructModelSpectrum(EeVsamples, spectrum)
  !---------------------------------------------------------------------

    implicit none 

    real(double), intent(in) :: EeVsamples(:)
    real(single), intent(out) :: spectrum(size(EeVsamples))

    !-------------------------------------------------------------------
    ! Start:

    spectrum = 1.0 ! can think about a more sophisticated model later!

  end subroutine constructModelSpectrum

  !---------------------------------------------------------------------
  function calcEeVFunctionNorm(EeVsamples, spectrum, filterTrans&
    , vigEfunctionAtSrc, qeArray, requiredCorrections) result(norm)
  !---------------------------------------------------------------------

    implicit none 

    real(double),   intent(in) :: EeVsamples(:)
    real(single),   intent(in) :: spectrum(         size(EeVsamples)),&
                                  filterTrans(      size(EeVsamples)),&
                                  vigEfunctionAtSrc(size(EeVsamples)),&
                                  qeArray(:,:)
    integer(int32), intent(in) :: requiredCorrections
    real(single)               :: norm

    real(single) :: totalQe(size(EeVsamples)), y(size(EeVsamples))
    integer :: eI, numESamples

    !-------------------------------------------------------------------
    ! Start:

    numESamples = size(EeVsamples)
    y = spectrum
    if (.not. btest(requiredCorrections, TREAT_FILTER_BIT))&
      y = y * filterTrans
    if (.not. btest(requiredCorrections, TREAT_VIGNET_BIT))&
      y = y * vigEfunctionAtSrc

    if (.not. btest(requiredCorrections, TREAT_QE_BIT)) then
      if (size(qeArray,2) .ne. numESamples)&
call error('bug0',"Bug in module exposure_map_pcms. Please contact the code developer.")

      totalQe = (/(sum(qeArray(:,eI)), eI=1,numESamples)/)
      y = y * totalQe
    end if

    norm = trapezoidSum(real(EeVsamples, single), y)
    ! in ssclib/array_utils.

  end function calcEeVFunctionNorm

  subroutine calcCcdExpMap(tempEventListSetName, eventSet, ccdNodeNum&
  , rawYLo, patternLo, patternHi, flagMask, EeVRanges&
  , timeAttWcs, pleaseCorrectDither, numMicropixelDivs, expInfo&
  , EeVsamples, EeVfunctions, qeArray, fluxCube, ccdExpMap, ccdFlatExp&
  , doDummy, regionTabName)
  !---------------------------------------------------------------------

    implicit none

    character(*),       intent(in)           :: tempEventListSetName
    type(DataSetT),     intent(in)           :: eventSet
    integer(int8),      intent(in)           :: ccdNodeNum
integer(int32),     intent(in)           :: patternLo, patternHi, flagMask, rawYLo
type(RangeT),       intent(in)           :: EeVRanges(:)
    type(localTAWtype), intent(in)           :: timeAttWcs
    logical(bool),      intent(in)           :: pleaseCorrectDither
    integer(int16),     intent(in)           :: numMicropixelDivs
    type(expInfoType),  intent(in)           :: expInfo
    real(single),       pointer,    optional :: fluxCube(:,:,:)
    real(double),       intent(in)           :: EeVsamples(:)
    real(single),       intent(in)           :: EeVfunctions(size(EeVsamples)),&
                                                qeArray(:,:)
    real(single),       intent(out)          :: ccdExpMap(:,:),&
                                                ccdFlatExp
    logical,      intent(in), optional :: doDummy
    character(*), optional, intent(in) :: regionTabName

    integer(int32) :: instrumentId
    logical :: localDoDummy

    !-------------------------------------------------------------------
    ! Start

    if (present(doDummy)) then
      localDoDummy = doDummy
    else
      localDoDummy = .false.
    end if

    call CAL_getState(instrument=instrumentId)

    select case(instrumentId)
    case(EMOS1, EMOS2)
!      call calcMosCcdExpMap(eventSet, ccdNodeNum, dssPointer&
!        , timeAttWcs, pleaseCorrectDither, numMicropixelDivs, expInfo&
!        , EeVsamples, EeVfunctions, qeArray, fluxCube, ccdExpMap, ccdFlatExp&
!        , localDoDummy)
      call calcMosCcdExpMap(tempEventListSetName, eventSet, ccdNodeNum&
        , rawYLo, patternLo, patternHi, flagMask, EeVRanges&
        , timeAttWcs, pleaseCorrectDither, numMicropixelDivs, expInfo&
        , EeVsamples, EeVfunctions, qeArray, fluxCube, ccdExpMap, ccdFlatExp&
        , localDoDummy, regionTabName)

    case(EPN)
!      call calcPnCcdExpMap(eventSet, ccdNodeNum, dssPointer&
!        , timeAttWcs, pleaseCorrectDither, numMicropixelDivs, expInfo&
!        , EeVsamples, EeVfunctions, qeArray, fluxCube, ccdExpMap, ccdFlatExp&
!        , localDoDummy)
      call calcPnCcdExpMap(tempEventListSetName, eventSet, ccdNodeNum&
        , rawYLo, patternLo, patternHi, flagMask, EeVRanges&
        , timeAttWcs, pleaseCorrectDither, numMicropixelDivs, expInfo&
        , EeVsamples, EeVfunctions, qeArray, fluxCube, ccdExpMap, ccdFlatExp&
        , localDoDummy, regionTabName)

    case default
call error('badInstrument',"The instrument is not recognized.")
    end select

  end subroutine calcCcdExpMap

  !---------------------------------------------------------------------
!  subroutine calcPnCcdExpMap(eventSet, ccdNodeNum, dssPointer&
!  , timeAttWcs, pleaseCorrectDither, numMicropixelDivs, expInfo&
!  , EeVsamples, EeVfunctions, qeArray, fluxCube, ccdExpMap, ccdFlatExp&
!  , doDummy)
  subroutine calcPnCcdExpMap(tempEventListSetName, eventSet, ccdNodeNum&
  , rawYLo, patternLo, patternHi, flagMask, EeVRanges&
  , timeAttWcs, pleaseCorrectDither, numMicropixelDivs, expInfo&
  , EeVsamples, EeVfunctions, qeArray, fluxCube, ccdExpMap, ccdFlatExp&
  , doDummy, regionTabName)
  !---------------------------------------------------------------------
  ! This subroutine takes the following inputs:
  !
  !	(i) the name of an event list;
  !	(ii) a list of dss component numbers;
  !	(iii) an array of samples of the expected relative flux of events as a function of energy (event list column PI), PATTERN and spatial coordinates (RAWX, RAWY, DETX, DETY, X, Y).
  !
  ! The subroutine returns a map (in RAWX,RAWY) of the predicted actual event count at each pixel. This is done by integrating the event-flux function, multiplied by the intrinsic EPIC exposure function, over the ranges in energy and pattern implicit in the dss.


  ! The steps in this process are as follows:
  !
  !	(i) From the input list of dss components, groups of components which share the same filters on energy and pattern are identified.
  !	(ii) Each of these combinations of energy and pattern filters may be found to define overlapping regions on the PI-PATTERN plane. This is not



!*** it is assumed that all dss components share the same filters on CCDNR and TIME.
!*** region or other 2d filters are allowed only between RAWX/RAWY, DETX/DETY and X/Y.


    implicit none

    character(*),       intent(in)           :: tempEventListSetName
    type(DataSetT),     intent(in)           :: eventSet
    integer(int8),      intent(in)           :: ccdNodeNum
!    type(DssT),         intent(in)           :: dssPointer
integer(int32),     intent(in)           :: patternLo, patternHi, flagMask, rawYLo
type(RangeT),       intent(in)           :: EeVRanges(:)
    type(localTAWtype), intent(in)           :: timeAttWcs
    logical(bool),      intent(in)           :: pleaseCorrectDither
    integer(int16),     intent(in)           :: numMicropixelDivs
    type(expInfoType),  intent(in)           :: expInfo
    real(single),       pointer,    optional :: fluxCube(:,:,:)
    real(double),       intent(in)           :: EeVsamples(:)
    real(single),       intent(in)           :: EeVfunctions(size(EeVsamples)),&
                                                qeArray(:,:)
    real(single),       intent(out)          :: ccdExpMap(:,:),&
                                                ccdFlatExp
    logical,      intent(in), optional :: doDummy
    character(*), optional, intent(in) :: regionTabName

    logical(bool) :: somePsfOnChip, xyDitherWasDone
    integer :: xSize, ySize
!    integer, pointer :: dssCompNums(:)
    integer(int8) :: patternNum
    integer(int16) ::&
      startXi, finisXi, startYi, finisYi,&
! maxPiPatComboIndex, piPatComboIndex,&
      xi, yi
    real(single), allocatable ::&
      ccdSpatialMap(:,:), tempArray(:,:)
!, ccdFlatExpMap(:,:)
!    integer(int16), allocatable :: uniquePiPatCombos(:)
    integer(int32) :: maxNumPatterns, numEeVSamples, eI
!, compIndex
!    logical(bool), allocatable :: compMask(:)
!    type(DScompT) :: dssComp
!    integer, allocatable :: dssCompNumsSubset(:)
!    type(RangeT), pointer :: piRanges(:)
!    type(RangeT), allocatable :: EeVRanges(:)
    real(single) :: ccdFlatExpTerm
integer(int8), pointer :: badPixelMap(:,:)
!type(DataSetT) :: eventSet
!logical(bool) :: evattIsSet(4)
!    character(8) :: exposuTabName
character(2) :: ccdStr
!    type(TableT) :: exposuTab
type(BitMaskT), allocatable :: flagBitMask(:)

logical :: testFlag = .false.
!character(6) :: ranStr
!type(TableT) :: eventTab
!real(single) :: dummyY(size(EeVsamples)), dummyT, dummyR, !dummyZ(size(EeVsamples))

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '    aaa calcPnCcdExpMap'

    call message(AppMsg, Verbose, '  Calculating the exposure map for &
      &this CCD:')

    if (present(doDummy)) then
      if (doDummy) then
        call warning('calcCcdExpMap', "This is a dummy routine! It doesn't &
          &return the real ccdExpMap.")
        ccdExpMap  = 1.0
        ccdFlatExp = 1.0
  return
      end if
    end if
if (testFlag) print *, ' bbb'

    numEeVSamples = size(EeVsamples)
    maxNumPatterns = size(qeArray, 1)
    if (size(qeArray,2) .ne. numEeVSamples)&
call error('bug6',"Please contact the code developer.")

    xSize = size(ccdExpMap, 1)
    ySize = size(ccdExpMap, 2)
    ccdExpMap  = 0.0
    ccdFlatExp = 0.0

    somePsfOnChip = .true. ! default
    if (present(fluxCube)) then
      if (size(fluxCube,1) .ne. numEeVSamples)&
call error('bug7',"Please contact the code developer.")

      startXi = max(1,     lbound(fluxCube, 2))
      finisXi = min(xSize, ubound(fluxCube, 2))
      startYi = max(1,     lbound(fluxCube, 3))
      finisYi = min(ySize, ubound(fluxCube, 3))

      if (startXi > finisXi .or. startYi > finisYi) then
        somePsfOnChip = .false.
      else
        do eI = 1, numEeVSamples

          fluxCube(eI,:,:) = fluxCube(eI,:,:) * EeVfunctions(eI)
        end do
      end if
    else
      startXi = 1
      finisXi = xSize
      startYi = 1
      finisYi = ySize
    end if
if (testFlag) print *, ' ccc'

!*** if (somePsfOnChip) calculate OOTEs. Make a chip map of ootes and multiply this in the same way as the flux cube. Then calculate the fraction of ootes not selected and return this.

    allocate(tempArray(startXi:finisXi, startYi:finisYi))
    call getPnBadPixelMap(eventSet, ccdNodeNum, badPixelMap)

    xyDitherWasDone = expInfo%randXY
    call readEvatts(eventSet)

!    call dssCompNumsWhichPassThisValue(dssPointer, 'CCDNR'&
!      , real(ccdNodeNum, single), dssCompNums)

!    ! Separate dss components into groups with the same PI-PATTERN filters:
!    allocate(uniquePiPatCombos(size(dssCompNums)))
!    call determineUniqueFilterCombos(dssPointer, dssCompNums&
!      , (/'PI     ', 'PATTERN'/), uniquePiPatCombos)
!if (testFlag) print *, ' ddd'

    allocate(ccdSpatialMap(xSize, ySize))
!    allocate(compMask(size(uniquePiPatCombos)))
!    maxPiPatComboIndex = maxval(uniquePiPatCombos)
!    do piPatComboIndex = 1, maxPiPatComboIndex
!    if (.not. any(uniquePiPatCombos == piPatComboIndex)) cycle

!      write(errstr, '(a, i3, a)') '    PI-PATTERN filter number '&
!        , piPatComboIndex,':'
!      call message(AppMsg, Verbose)

!      compMask = (uniquePiPatCombos == piPatComboIndex)
!      compIndex = smallestTrueIndex(compMask)
!      ! smallestTrueIndex is in ssclib/array_utils.

!      dssComp = dssComponent(dssPointer, dssCompNums(compIndex))
!      ! The first dss component which has this PI-PATTERN filter combination.

!      ! Get a list of the dss component numbers for this PI-PATTERN filter
!      ! combination:
!      allocate(dssCompNumsSubset(count(compMask)))
!      dssCompNumsSubset = pack(dssCompNums, compMask)

!if (testFlag) print *, 'This PI-PATTERN combo has dss component numbers'&
!  , dssCompNumsSubset

allocate(flagBitMask(1))
flagBitMask%onBits = 0
flagBitMask%offBits = flagMask

      do patternNum = 0, maxNumPatterns - 1
!      if (.not. valueIsPassedByDssComp(dssComp, 'PATTERN'&
!        , real(patternNum, single))) cycle
!      ! valueIsPassedByDssComp is in ssclib/dss_aux.
if (patternNum < patternLo .or. patternNum > patternHi) cycle

        write(errstr, '(a, i3, a)') '      PATTERN number ',patternNum,':'
        call message(AppMsg, Verbose)

if (testFlag) print *, ' eee'
!        call calcCcdSpatialMap(ccdNodeNum, patternNum&
!          , dssPointer, dssCompNumsSubset, timeAttWcs, badPixelMap&
!          , numMicropixelDivs, xyDitherWasDone, pleaseCorrectDither&
!          , ccdSpatialMap)

        call calcCcdSpatialMap(tempEventListSetName, ccdNodeNum, patternNum&
          , rawYLo, flagBitMask, timeAttWcs, badPixelMap&
          , numMicropixelDivs, xyDitherWasDone, pleaseCorrectDither&
          , ccdSpatialMap, regionTabName)
if (testFlag) print *, ' fff'

      if (all(ccdSpatialMap <= 0.0)) cycle
if (testFlag) print *, ' fff0'

if (testFlag) then
!  ranStr = randomLetterStr('000000')
!  print *, 'random str is ', ranStr
  write(ccdStr, '(i2.2)') patternNum
  call dumpImageToFits(ccdSpatialMap&
!!    , 'ccdSpatialMap_'//ccdStr//'_'//ranStr//'.ds')
, 'spatialMap_pattern'//ccdStr//'.ds')
end if
if (testFlag) print *, ' fff1'

!        call extractSliceRanges(dssPointer, dssCompNumsSubset, 'PATTERN', 'PI'&
!          , real(patternNum, single), piRanges)
!if (testFlag) print *, ' ggg'

!        allocate(EeVranges(size(piRanges)))
!        EeVranges = convertPiToEeV(piRanges)
!        deallocate(piRanges)
        ccdFlatExpTerm = integrateRanges(real(EeVSamples, single)&
          , EeVfunctions * qeArray(patternNum + 1, :), EeVRanges)
if (testFlag) print *, ' hhh'

        if (somePsfOnChip) then
          if (present(fluxCube)) then
            do xi = startXi, finisXi
              do yi = startYi, finisYi
                tempArray(xi,yi) = integrateRanges(real(EeVSamples, single)&
                    , fluxCube(:,xi,yi) * qeArray(patternNum + 1, :), EeVRanges)
              end do
            end do
          else
            tempArray = ccdFlatExpTerm
          end if
if (testFlag) print *, ' iii'

          ccdExpMap(startXi:finisXi, startYi:finisYi)&
            = ccdExpMap(startXi:finisXi, startYi:finisYi)&
            + ccdSpatialMap(startXi:finisXi, startYi:finisYi) * tempArray
        end if
if (testFlag) print *, ' jjj'

        ccdFlatExp = ccdFlatExp + ccdFlatExpTerm * sum(ccdSpatialMap)
if (testFlag) print *, ' kkk'

!        deallocate(EeVRanges)
if (testFlag) exit
      end do
!      deallocate(dssCompNumsSubset)
!    end do
if (testFlag) print *, ' lll'

!    deallocate(dssCompNums)
    deallocate(badPixelMap)

    call message(AppMsg, Noisy, '  Leaving subroutine calcPnCcdExpMap.')

if (testFlag) print *, ' zzz calcPnCcdExpMap'
  end subroutine calcPnCcdExpMap

  !---------------------------------------------------------------------
  subroutine getPnBadPixelMap(eventSet, ccdNodeNum, badPixelMap)
  !---------------------------------------------------------------------
  ! NOTE! The calling program should deallocate the pointer badPixelMap after use.

    implicit none

    type(DataSetT), intent(in) :: eventSet
    integer(int8),  intent(in) :: ccdNodeNum
    integer(int8),  pointer    :: badPixelMap(:,:)

    integer(int16) ::&
      xSize, ySize, xi, yiStart, yiFinis, windowMinRawX, windowMinRawY,&
      windowRawDX, windowRawDY, windowMaxRawX, windowMaxRawY
    character(2) :: ccdStr
    character(8) ::&
      badPixTabName, exposuTabName, oldOffsetTabName, offsetTabName
    type(TableT) :: badPixTab, offsetsTab, exposuTab
    type(ColumnT) ::&
      badXcol, badYcol, badYextCol, badTypeCol, badFlagCol
    integer(int16), pointer, dimension(:) ::&
      badX, badY, badYext, badType, badFlag, offsetRawX, offsetPha
    integer(int32) :: numBadpixRows, row, numOffsetsRows
    integer(int8), pointer :: offsetCcd(:)
    integer(int16), parameter :: pnOffsetLimit = 0
    integer(int16), pointer :: fovMap(:,:)
    integer :: bit
    logical :: doOffset

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, '    Reading the BADPIX and OFFSETS &
      &extensions and WINDOW keywords from the event list.')

    xSize = EPN_MAX_X
    ySize = EPN_MAX_Y
    allocate(badPixelMap(0:xSize+1,0:ySize+1))
    badPixelMap = 0

    write(ccdStr, '(i2.2)') ccdNodeNum
    badPixTabName    = 'BADPIX'//ccdStr
    exposuTabName    = 'EXPOSU'//ccdStr
    oldOffsetTabName = 'OFFSET'//ccdStr

    !...................................................................
    ! Bad pixels first:

    badPixTab = table(eventSet, badPixTabName)
    badXcol    = column(badPixTab, "RAWX")
    badYcol    = column(badPixTab, "RAWY")
    badYextCol = column(badPixTab, "YEXTENT")
    badTypeCol = column(badPixTab, "TYPE")
    badFlagCol = column(badPixTab, "BADFLAG")

    badX    => int16Data(badXcol)
    badY    => int16Data(badYcol)
    badYext => int16Data(badYextCol)
    badType => int16Data(badTypeCol)
    ! Type: 0=OK, 1=hot, 2=flickering, 3=dead, 4=pinhole, 5=anomalous
    badFlag => int16Data(badFlagCol)
    ! Flag: 1=uplinked, 2=in CCF but not uplinked, 3=new

    numBadpixRows = numberOfRows(badPixTab)
    do row = 0, numBadpixRows - 1
      xi = badX(row)
      yiStart = badY(row)
      yiFinis = yiStart + badYext(row) - 1

      ! Check that no bad pixel has invalid coordinates (RAWX=0 is valid)

      if (xi < 0 .or. yiStart < 1 .or. xi > xSize+1 .or. yiFinis > ySize) then
        write(errstr,*) "Invalid bad pixel coordinates x=", xi, ", y from "&
          , yiStart, " to ", yiFinis, " at row", row
        call warning("badBadPixCoords")
    cycle
      end if

      select case(int(badType(row), int32))
      case (INTACT)
    cycle

      case (HOT)
        if (badFlag(row) > 1) then
          bit = BRIGHTPIX_BIT
        else
          bit = ONBOARD_BADPIX_BIT
        endif

      case (DEAD)
        bit = DEADPIX_BIT

      case default
call error('badBadPixType',"The value of the TYPE column of the "&
  //trim(badPixTabName)//" table was not recognized.")
      end select

      if (xi > 0) then
        badPixelMap(xi, yiStart:yiFinis)&
          = ibset(badPixelMap(xi, yiStart:yiFinis), bit)
      end if
    end do

    call release(badPixTab)

    !...................................................................
    ! Offsets:

    doOffset = .false. ! default
    if (hasBlock(eventSet, 'OFFSETS')) then
      doOffset = .true.
      offsetTabName = 'OFFSETS'
    else if (hasBlock(eventSet, oldOffsetTabName)) then
      call warning('obsoleteOffsetExtensionName',"")
      doOffset = .true.
      offsetTabName = oldOffsetTabName
    end if

    if (doOffset) then
      offsetsTab = table(eventSet, offsetTabName)
      if (numberOfRows(offsetsTab) > 0) then
        offsetRawX => int16Data(column(offsetsTab, 'RAWX'))
        offsetPha  => int16Data(column(offsetsTab, 'OFFSETX'))
        offsetCcd  => int8Data( column(offsetsTab, 'CCDNR'))

        numOffsetsRows = numberOfRows(offsetsTab)
        do row = 0, numOffsetsRows - 1
          if (offsetCcd(row) .ne. ccdNodeNum&
          .or. offsetRawX(row) < 0 .or. offsetRawX(row) > xSize+1)&
        cycle

          if (offsetPha(row) > pnOffsetLimit)&
            badPixelMap(offsetRawX(row),:)&
              = ibset(badPixelMap(offsetRawX(row),:), BAD_OFFSET_BIT)
        end do
      end if

      call release(offsetsTab)
    end if

    !...................................................................
    ! Get window borders:

    exposuTab = table(eventSet, exposuTabName)
    windowMinRawX = int16attribute(exposuTab, 'WINDOWX0')
    windowMinRawY = int16attribute(exposuTab, 'WINDOWY0')
    windowRawDX   = int16attribute(exposuTab, 'WINDOWDX')
    windowRawDY   = int16attribute(exposuTab, 'WINDOWDY')
    call release(exposuTab)

    windowMaxRawX = min(xSize,   windowMinRawX + windowRawDX - 1_int16)
    windowMaxRawY = min(ySize,   windowMinRawY + windowRawDY - 1_int16)
    windowMinRawX = max(1_int16, windowMinRawX)
    windowMinRawY = max(1_int16, windowMinRawY)

    badPixelMap = ibset(badPixelMap, OUTSIDE_WINDOW_BIT)
    badPixelMap(windowMinRawX:windowMaxRawX, windowMinRawY:windowMaxRawY)&
      = ibclr(badPixelMap(windowMinRawX:windowMaxRawX&
      , windowMinRawY:windowMaxRawY), OUTSIDE_WINDOW_BIT)

    !...................................................................
    ! Out-of-FOV pixels:

    call CAL_getFOVmap(fovMap)
    if (size(fovMap,1) .ne. xSize .or. size(fovMap,2) .ne. ySize) then
call error('bug8',"Please contact the code developer.")
    end if

    where(fovMap == 0)&
      badPixelMap(1:xSize,1:ySize) = ibset(badPixelMap(1:xSize,1:ySize)&
      , OUTSIDE_FOV_BIT)

    call CAL_releaseMemory(fovMap)

  end subroutine getPnBadPixelMap

  !---------------------------------------------------------------------
!  subroutine calcMosCcdExpMap(eventSet, ccdNodeNum, dssPointer&
!  , timeAttWcs, pleaseCorrectDither, numMicropixelDivs, expInfo&
!  , EeVsamples, EeVfunctions, qeArray, fluxCube, ccdExpMap, ccdFlatExp&
!  , doDummy)
  subroutine calcMosCcdExpMap(tempEventListSetName, eventSet, ccdNodeNum&
  , rawYLo, patternLo, patternHi, flagMask, EeVRanges&
  , timeAttWcs, pleaseCorrectDither, numMicropixelDivs, expInfo&
  , EeVsamples, EeVfunctions, qeArray, fluxCube, ccdExpMap, ccdFlatExp&
  , doDummy, regionTabName)
  !---------------------------------------------------------------------
  ! This subroutine takes the following inputs:
  !
  !	(i) the name of an event list;
  !	(ii) a list of dss component numbers;
  !	(iii) an array of samples of the expected relative flux of events as a function of energy (event list column PI), PATTERN and spatial coordinates (RAWX, RAWY, DETX, DETY, X, Y).
  !
  ! The subroutine returns a map (in RAWX,RAWY) of the predicted actual event count at each pixel. This is done by integrating the event-flux function, multiplied by the intrinsic EPIC exposure function, over the ranges in energy and pattern implicit in the dss.

    implicit none

    character(*),       intent(in)           :: tempEventListSetName
    type(DataSetT),     intent(in)           :: eventSet
    integer(int8),      intent(in)           :: ccdNodeNum
!    type(DssT),         intent(in)           :: dssPointer
integer(int32),     intent(in)           :: patternLo, patternHi, flagMask, rawYLo
type(RangeT),       intent(in)           :: EeVRanges(:)
    type(localTAWtype), intent(in)           :: timeAttWcs
    logical(bool),      intent(in)           :: pleaseCorrectDither
    integer(int16),     intent(in)           :: numMicropixelDivs
    type(expInfoType),  intent(in)           :: expInfo
    real(single),       pointer,    optional :: fluxCube(:,:,:)
    real(double),       intent(in)           :: EeVsamples(:)
    real(single),       intent(in)           :: EeVfunctions(size(EeVsamples)),&
                                                qeArray(:,:)
    real(single),       intent(out)          :: ccdExpMap(:,:),&
                                                ccdFlatExp
    logical,      intent(in), optional :: doDummy
    character(*), optional, intent(in) :: regionTabName

    logical(bool) :: somePsfOnChip, xyDitherWasDone
    integer :: xSize, ySize
!    integer, pointer :: dssCompNums(:)
    integer(int8) :: patternNum
    integer(int16) :: startXi, finisXi, startYi, finisYi, xi, yi
    real(single), allocatable :: ccdSpatialMap(:,:), tempArray(:,:)
!    integer(int16), allocatable :: uniquePiPatCombos(:)
!    type(boxChainType), pointer :: listOfBoxChains(:)
!    type(boxIndexChainType) :: boxIndexChain
    integer(int32) ::&
      eI, numEeVsamples,&
      maxNumPatterns
!, numBoxChains, boxChainNum,numBoxXEdges, formerIndex, numNewEeVSamples
!    integer(int16), allocatable :: EeVSampleIndices(:), boxEdgeIndices(:)
!    real(double), pointer :: mergedEeVSamples(:)
!    type(StepYT) :: totalSelectedQeStepEeV
!    type(StepYT), pointer ::&
!      stepEFluxCube(:,:,:), stepEeVfunctions(:), stepEQeArray(:,:)
    real(single) ::&
      transposedQeArray(size(qeArray, 2),size(qeArray, 1)), ccdFlatExpTerm
    integer(int8), pointer :: badPixelMap(:,:)
type(BitMaskT), allocatable :: flagBitMask(:)

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, 'aaa calcMosCcdExpMap'

    call message(AppMsg, Verbose, '  Calculating the exposure map for &
      &this CCD:')

    if (present(doDummy)) then
      if (doDummy) then
        call warning('calcCcdExpMap', "This is a dummy routine! It doesn't &
          &return the real ccdExpMap.")
        ccdExpMap  = 1.0
        ccdFlatExp = 1.0
  return
      end if
    end if
if (testFlag) print *, 'bbb'

    numEeVSamples = size(EeVsamples)
    maxNumPatterns = size(qeArray, 1)
    if (size(qeArray,2) .ne. numEeVSamples)&
call error('bug9',"Please contact the code developer.")

    xSize = size(ccdExpMap, 1)
    ySize = size(ccdExpMap, 2)
    ccdExpMap  = 0.0
    ccdFlatExp = 0.0

    transposedQeArray = transpose(qeArray)
if (testFlag) print *, 'ccc'

    somePsfOnChip = .true. ! default
    if (present(fluxCube)) then
      if (size(fluxCube,1) .ne. numEeVSamples)&
call error('bug10',"Please contact the code developer.")

      startXi = max(1,     lbound(fluxCube, 2))
      finisXi = min(xSize, ubound(fluxCube, 2))
      startYi = max(1,     lbound(fluxCube, 3))
      finisYi = min(ySize, ubound(fluxCube, 3))

      if (startXi > finisXi .or. startYi > finisYi) then
        somePsfOnChip = .false.
      else
        do eI = 1, numEeVSamples
          fluxCube(eI,:,:) = fluxCube(eI,:,:) * EeVfunctions(eI)
        end do
      end if
    else
      startXi = 1
      finisXi = xSize
      startYi = 1
      finisYi = ySize
    end if
if (testFlag) print *, 'ddd'

    allocate(tempArray(startXi:finisXi, startYi:finisYi))

if (testFlag) print *, 'ddda'
    call getMosBadPixelMap(eventSet, ccdNodeNum, badPixelMap)
if (testFlag) print *, 'eee'

    xyDitherWasDone = expInfo%randXY

if (testFlag) print *, 'fff'

    call readEvatts(eventSet)
if (testFlag) print *, 'ggg'

if (testFlag) print *, 'hhh'

!    call dssCompNumsWhichPassThisValue(dssPointer, 'CCDNR'&
!      , real(ccdNodeNum, single), dssCompNums)
if (testFlag) print *, 'iii'

!    if (size(dssCompNums) <= 0)&
!call error('noCcdsPassDss',"The data subspace passes no value of CCDNR.")

!    ! Separate dss components into groups with the same PI-PATTERN filters:
!    allocate(uniquePiPatCombos(size(dssCompNums)))
!    call determineUniqueFilterCombos(dssPointer, dssCompNums&
!      , (/'PI     ', 'PATTERN'/), uniquePiPatCombos) ! in ssclib/dss_aux.
if (testFlag) print *, 'jjj'

    !...................................................................
    ! We want now to get rid of overlaps between different PI-PATTERN
    ! filters. The best way to do this is to treat each of these
    ! filters as a chain or series of rectangular regions in PI-PATTERN
    ! space defined by independent sets of ranges in PI and PATTERN.
    ! The overlaps between the rectangles in each filter are
    ! found, and new sets of filter chains are constructed, using the
    ! dss_aux routine calcBoxChainOverlaps().

!    call calcNonOverlappingBoxChains(dssPointer, dssCompNums&
!      , uniquePiPatCombos, 'PI', 'PATTERN', listOfBoxChains)
    ! in ssclib/boxchains.
if (testFlag) print *, 'kkk'

    !...................................................................
    ! Now it is time to do the integrations. For each of the non-
    ! overlapping EEV-PATTERN filters that are at present stored in the
    ! pointer array listOfBoxChains, we must first get a ccd pixel
    ! map of those pixels that pass the corresponding spatial filter;
    ! then we must integrate all the EEV- and PATTERN-dependent functions
    ! over the relevant EEV-PATTERN ranges.

    allocate(ccdSpatialMap(xSize, ySize))
!    numBoxChains = size(listOfBoxChains)

!    do boxChainNum = 1, numBoxChains
!      write(errstr, '(a, i3, a)') '    PI-PATTERN filter number '&
!        , boxChainNum, ':'
!      call message(AppMsg, Verbose)

!if (testFlag) print *, 'This PI-PATTERN combo has dss component numbers'&
!  , listOfBoxChains(boxChainNum)%dssCompNums

!      call calcCcdSpatialMap(ccdNodeNum, 0_int8&
!        , dssPointer, listOfBoxChains(boxChainNum)%dssCompNums, timeAttWcs&
!        , badPixelMap, numMicropixelDivs, xyDitherWasDone, pleaseCorrectDither&
!        , ccdSpatialMap)

allocate(flagBitMask(1))
flagBitMask%onBits = 0
flagBitMask%offBits = flagMask

      call calcCcdSpatialMap(tempEventListSetName, ccdNodeNum, 0_int8&
        , rawYLo, flagBitMask, timeAttWcs&
        , badPixelMap, numMicropixelDivs, xyDitherWasDone, pleaseCorrectDither&
        , ccdSpatialMap, regionTabName)
      ! patternNum=0_int8 for mos because there is no dependence of the
      ! spatial map on pattern.

if (testFlag) print *, 'lll'
!    if (all(ccdSpatialMap <= 0.0)) cycle
!if (all(ccdSpatialMap <= 0.0)) call error('spatialMapAllZero',"")

!      ! At the moment we have one list of energies which are the energy samples in EeVSamples, plus another which represents the ensemble of bounding energies for the PI-energy boxes in boxChain (note that some of the latter may still be unbounded). Now what we want to do is to generate two matching lists of indices to a common or merged list of energies: 
!
!      call convertBoxesToIndexStyle(listOfBoxChains(boxChainNum), boxIndexChain)
!if (testFlag) print *, 'mmm'
!      numBoxXEdges = size(boxIndexChain%xEdges)
!if (testFlag) print *, 'nnn'
!
!      boxIndexChain%xEdges = convertPiToEeV(boxIndexChain%xEdges)
!      boxIndexChain%yEdges = convertPiToEeV(boxIndexChain%yEdges)
!if (testFlag) print *, 'ooo'
!
!      if (boxIndexChain%xEdges(1)%type == UNDEFINED) then
!        boxIndexChain%xEdges(1)%type = INCLUSIVE
!        boxIndexChain%xEdges(1)%value = EeVSamples(1)
!      end if
!      if (boxIndexChain%xEdges(numBoxXEdges)%type == UNDEFINED) then
!        boxIndexChain%xEdges(numBoxXEdges)%type = INCLUSIVE
!        boxIndexChain%xEdges(numBoxXEdges)%value = EeVSamples(numEeVSamples)
!      end if
!
!      allocate(EeVSampleIndices(numEeVSamples))
!      allocate(boxEdgeIndices(numBoxXEdges))
!      call generateMergedListWithIndices(EeVSamples&
!        , boxIndexChain%xEdges%value, mergedEeVSamples&
!        , EeVSampleIndices, boxEdgeIndices)
!
!      do eI = 1, size(boxIndexChain%boxes)
!        formerIndex = boxIndexChain%boxes(eI)%x%lowerI
!        boxIndexChain%boxes(eI)%x%lowerI = boxEdgeIndices(formerIndex)
!        formerIndex = boxIndexChain%boxes(eI)%x%upperI
!        boxIndexChain%boxes(eI)%x%upperI = boxEdgeIndices(formerIndex)
!      end do
!
!      numNewEeVSamples = size(mergedEeVSamples)
!if (testFlag) print *, 'ppp'
!
!      !..................................................................
!      ! Now convert the flux cube to step-function type, interpolating missing energy values:
!
!      if (present(fluxCube) .and. somePsfOnChip) then
!        call convertAndInterpolate(fluxCube, mergedEeVSamples&
!          , EeVSampleIndices, stepEFluxCube)
!      end if
!
!      call convertAndInterpolate(EeVfunctions, mergedEeVSamples&
!        , EeVSampleIndices, stepEeVfunctions)
!
!      call convertAndInterpolate(transposedQeArray, mergedEeVSamples&
!        , EeVSampleIndices, stepEQeArray)
!      ! Using transposedQeArray here because it has energy as the first index.
!if (testFlag) print *, 'qqq'
!
!      !..................................................................
!      ! Integrate QE over the pattern selection:
!
!     do eI = 1, numNewEeVSamples
!if (testFlag) print *, '  qqqa eI:', eI
!        ! Integrate QE in the pattern direction across all the DSS filter boxes:
!
!        call initializeStepY(totalSelectedQeStepEeV)
!if (testFlag) print *, '  qqqb'
!        do patternNum = 0, maxNumPatterns - 1
!          totalSelectedQeStepEeV = addStepYVals(&
!            totalSelectedQeStepEeV,&
!            multiplyStepYVals(&
!              stepEQeArray(eI, patternNum + 1),&
!              boxIndexStepMask(boxIndexChain, eI, real(patternNum, single))&
!            )&
!          )
!        end do
!if (testFlag) print *, '  qqqc'
!
!        stepEeVfunctions(eI) = multiplyStepYVals(stepEeVfunctions(eI)&
!          , totalSelectedQeStepEeV)
!if (testFlag) print *, '  qqqd'
!
!        if (present(fluxCube) .and. somePsfOnChip) then
!          stepEFluxCube(eI,:,:) = multiplyStepYVals(stepEFluxCube(eI,:,:)&
!            , totalSelectedQeStepEeV)
!if (testFlag) print *, '  qqqe'
!        end if
!if (testFlag) print *, ''
!      end do
!if (testFlag) print *, 'rrr'
!
!      ccdFlatExpTerm = stepIntegrate(mergedEeVSamples&
!        , stepEeVfunctions)
!
!      if (somePsfOnChip) then
!        if (present(fluxCube)) then
!          tempArray = stepIntegrate(mergedEeVSamples&
!!                , stepEFluxCube)
!                , stepEFluxCube(:,startXi:finisXi, startYi:finisYi))
!        else
!          tempArray = ccdFlatExpTerm
!        end if
!
!        ccdExpMap(startXi:finisXi, startYi:finisYi)&
!          = ccdExpMap(startXi:finisXi, startYi:finisYi)&
!          + ccdSpatialMap(startXi:finisXi, startYi:finisYi) * tempArray
!      end if
!if (testFlag) print *, 'sss'
!
!      ccdFlatExp = ccdFlatExp + ccdFlatExpTerm * sum(ccdSpatialMap)
!if (testFlag) print *, 'ttt'
!
!      deallocate(listOfBoxChains(boxChainNum)%boxes)
!if (testFlag) print *, 'uuu'
!      deallocate(listOfBoxChains(boxChainNum)%dssCompNums)
!if (testFlag) print *, 'vvv'
!      deallocate(mergedEeVSamples, stepEeVfunctions, stepEQeArray)
!if (testFlag) print *, 'www'
!      if (present(fluxCube) .and. somePsfOnChip) deallocate(stepEFluxCube)
!if (testFlag) print *, 'xxx'
!    end do ! boxChainNum
!if (testFlag) print *, 'yyy'
!
!    call message(AppMsg, Noisy, '  Leaving subroutine calcMosCcdExpMap.')
!
!    deallocate(uniquePiPatCombos, listOfBoxChains)
!deallocate(badPixelMap)!, evattIsSet)
!!call release(eventSet)

if (.not.all(ccdSpatialMap <= 0.0)) then

      do patternNum = 0, maxNumPatterns - 1
      if (patternNum < patternLo .or. patternNum > patternHi) cycle
        write(errstr, '(a, i3, a)') '      PATTERN number ',patternNum,':'
        call message(AppMsg, Noisy)

        ccdFlatExpTerm = integrateRanges(real(EeVSamples, single)&
          , EeVfunctions * qeArray(patternNum + 1, :), EeVRanges)

        if (somePsfOnChip) then
          if (present(fluxCube)) then
            do xi = startXi, finisXi
              do yi = startYi, finisYi
                tempArray(xi,yi) = integrateRanges(real(EeVSamples, single)&
                    , fluxCube(:,xi,yi) * qeArray(patternNum + 1, :), EeVRanges)
              end do
            end do
          else
            tempArray = ccdFlatExpTerm
          end if

          ccdExpMap(startXi:finisXi, startYi:finisYi)&
            = ccdExpMap(startXi:finisXi, startYi:finisYi)&
            + ccdSpatialMap(startXi:finisXi, startYi:finisYi) * tempArray
        end if

        ccdFlatExp = ccdFlatExp + ccdFlatExpTerm * sum(ccdSpatialMap)

      end do
end if

    deallocate(badPixelMap)

if (testFlag) print *, 'zzz calcMosCcdExpMap'

  end subroutine calcMosCcdExpMap

  !---------------------------------------------------------------------
  subroutine getMosBadPixelMap(eventSet, ccdNodeNum, badPixelMap)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in) :: eventSet
    integer(int8),  intent(in) :: ccdNodeNum
    integer(int8),  pointer    :: badPixelMap(:,:)

    integer(int16) ::&
      xSize, ySize, xi, yiStart, yiFinis, windowMinRawX, windowMinRawY,&
      windowRawDX, windowRawDY, windowMaxRawX, windowMaxRawY
    character(2) :: ccdStr
    character(8) :: badPixTabName, exposuTabName
    type(TableT) :: badPixTab, exposuTab
    type(ColumnT) ::&
      badXcol, badYcol, badYextCol, badTypeCol, badFlagCol
    integer(int16), pointer, dimension(:) ::&
      badX, badY, badYext, badType, badFlag
    integer(int32) :: numBadpixRows, row
    integer(int16), pointer :: fovMap(:,:)
    integer :: bit
    integer :: offsetX(-10:EMOS_MAX_X+20), offsetY(-10:EMOS_MAX_Y+20)
    logical(bool), allocatable :: tempMask(:,:)

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, '    Reading the BADPIX and OFFSETS &
      &extensions and WINDOW keywords from the event list.')

    xSize = EMOS_MAX_X
    ySize = EMOS_MAX_Y
    allocate(badPixelMap(0:xSize+1,0:ySize+1))
    badPixelMap = 0

    write(ccdStr, '(i2.2)') ccdNodeNum
    badPixTabName = 'BADPIX'//ccdStr
    exposuTabName = 'EXPOSU'//ccdStr

    !...................................................................
    ! Bad pixels first:

    badPixTab = table(eventSet, badPixTabName)
    badXcol    = column(badPixTab, "RAWX")
    badYcol    = column(badPixTab, "RAWY")
    badYextCol = column(badPixTab, "YEXTENT")
    badTypeCol = column(badPixTab, "TYPE")
    badFlagCol = column(badPixTab, "BADFLAG")

    badX    => int16Data(badXcol)
    badY    => int16Data(badYcol)
    badYext => int16Data(badYextCol)
    badType => int16Data(badTypeCol)
    ! Type: 0=OK, 1=hot, 2=flickering, 3=dead, 4=pinhole, 5=anomalous
    badFlag => int16Data(badFlagCol)
    ! Flag: 1=uplinked, 2=in CCF but not uplinked, 3=new

    numBadpixRows = numberOfRows(badPixTab)
    do row = 0, numBadpixRows - 1
      xi = badX(row)
      yiStart = badY(row)
      yiFinis = yiStart + badYext(row) - 1

      ! Check that no bad pixel has invalid coordinates (RAWX=0 is valid)

      if (xi < 0 .or. yiStart < 1 .or. xi > xSize+1 .or. yiFinis > ySize) then
        write(errstr,*) "Invalid bad pixel coordinates x=", xi, ", y from "&
          , yiStart, " to ", yiFinis, " at row", row
        call warning("badBadPixCoords")
    cycle
      end if

      select case(int(badType(row), int32))
      case (INTACT)
    cycle
!*** check that this is right for mos.

      case (DEAD)
        bit = DEADPIX_BIT

      case default
        if (badFlag(row) > 1) then
          bit = BRIGHTPIX_BIT
        else
          bit = ONBOARD_BADPIX_BIT
        endif
      end select

      if (xi > 0) then
        badPixelMap(xi, yiStart:yiFinis)&
          = ibset(badPixelMap(xi, yiStart:yiFinis), bit)
      end if
    end do

    call release(badPixTab)

    !...................................................................
    ! Offsets:

    call readBadOffsets(eventSet, offsetX, offsetY, int(ccdNodeNum))
    ! in emsaplib/badpixutils_mod.

    if (.not. (all(offsetX <= 0) .and. all(offsetY <= 0))) then
      allocate(tempMask(0:xSize+1,0:ySize+1))
      tempMask = (spread(offsetX(0:xSize+1), 2, ySize+2) > 0)&
            .or. (spread(offsetY(0:ySize+1), 1, xSize+2) > 0)
!*** HUH?? But offsetX for example is longer than xSize+2!!

      where(tempMask) badPixelMap = ibset(badPixelMap , BAD_OFFSET_BIT)
    end if

    !...................................................................
    ! Get window borders:

    exposuTab = table(eventSet, exposuTabName)
    windowMinRawX = int16attribute(exposuTab, 'WINDOWX0')
    windowMinRawY = int16attribute(exposuTab, 'WINDOWY0')
    windowRawDX   = int16attribute(exposuTab, 'WINDOWDX')
    windowRawDY   = int16attribute(exposuTab, 'WINDOWDY')
    call release(exposuTab)

    windowMaxRawX = min(xSize,   windowMinRawX + windowRawDX - 1_int16)
    windowMaxRawY = min(ySize,   windowMinRawY + windowRawDY - 1_int16)
    windowMinRawX = max(1_int16, windowMinRawX)
    windowMinRawY = max(1_int16, windowMinRawY)

    badPixelMap = ibset(badPixelMap, OUTSIDE_WINDOW_BIT)
    badPixelMap(windowMinRawX:windowMaxRawX, windowMinRawY:windowMaxRawY)&
      = ibclr(badPixelMap(windowMinRawX:windowMaxRawX&
      , windowMinRawY:windowMaxRawY), OUTSIDE_WINDOW_BIT)

    if (windowMaxRawX == xSize/2) then ! This criterion is the same
    ! as used by Jean Ballet in emenergy.

      badPixelMap(xSize/2, 1:ySize) = ibset(badPixelMap(xSize/2, 1:ySize)&
        , NODE_BOUNDARY_BIT)
    end if

    !...................................................................
    ! Out-of-FOV pixels:

    call CAL_getFOVmap(fovMap)
    if (size(fovMap,1) .ne. xSize .or. size(fovMap,2) .ne. ySize)&
call error('bug11',"Please contact the code developer.")

    where(fovMap == 0 .and. .not. btest(badPixelMap(1:xSize,1:ySize)&
      , OUTSIDE_WINDOW_BIT))
    ! For mos, only events within the window can get flagged OOFOV.

      badPixelMap(1:xSize,1:ySize) = ibset(badPixelMap(1:xSize,1:ySize)&
        , OUTSIDE_FOV_BIT)
    end where

    call CAL_releaseMemory(fovMap)

  end subroutine getMosBadPixelMap

  !---------------------------------------------------------------------
  subroutine calcCcdSpatialMap(tempEventListSetName, ccdNodeNum&
!  , dssPointer, dssCompNums, timeAttWcs, badPixelMap, numDivs&
  , patternNum, rawYLo, flagBitMask, timeAttWcs, badPixelMap, numDivs&
  , xyDitherWasDone, pleaseCorrectDither, ccdSpatialMap&
, regionTabName)
  !---------------------------------------------------------------------

  !
  !---------------------------------------------------------------------
  ! NOTE:
  !	(i) It is required that the cal is set up with the appropriate
  ! instrument, ccdChipId, ccdNodeId and ccdModeId beforehand.
  !	(ii) This routine calls CAL_SetState(randomize=.false.).
  !	(iii) The input argument 'wcs' should be the wcs of the X and
  ! Y columns of the event list.
  !---------------------------------------------------------------------

!*** see if there are components with the same region spatial filtering??
!*** special treatment if there is no region spatial filtering?

    implicit none

    character(*),       intent(in)  :: tempEventListSetName
    integer(int8),      intent(in)  :: ccdNodeNum, patternNum
integer(int32),     intent(in)  :: rawYLo
!    type(DssT),         intent(in)  :: dssPointer
!    integer(int32),     intent(in)  :: dssCompNums(:)
type(BitMaskT),     intent(in)  :: flagBitMask(:)
    type(localTAWtype), intent(in)  :: timeAttWcs
    integer(int8),       pointer    :: badPixelMap(:,:)
    integer(int16),     intent(in)  :: numDivs
    logical(bool),      intent(in)  :: xyDitherWasDone, pleaseCorrectDither
    real(single),       intent(out) :: ccdSpatialMap(:,:)
    character(*), optional, intent(in) :: regionTabName

    character(2) :: ccdStr
!    type(DScompT) :: dssComp
    type(DataSetT) :: set
    integer(int16) :: numDivsSquared, xsize, ysize, arrayShape(2), index
    integer(int32) ::&
      numCcdPixels, vectorShape(1), numRowsSel, numEdgePixels,&
      coarseRow, i, numFineRows, fineRow
!, numDssComps, compNum
    real(single) :: realNumDivsSquared
    type(TableT) :: coarseTab, fineTab
    type(ColumnT) ::&
      flagCol, weightCol, rawFlagCol, fineFlagCol, rowInCoarseCol
    integer(int32), pointer ::&
      flag(:), rawFlag(:), fineFlag(:), rowInCoarse(:)
    real(single), pointer ::&
      weight(:)
    character(255) :: selExpression
    logical(bool), allocatable, dimension(:,:) ::&
      selArray, figureEdges, groundEdges, totalEdgeMap
    logical(bool), pointer :: flaggedCcdMask(:,:)
    integer(int32), allocatable :: finalFlags(:)

!character(30) :: dumpName
!integer :: j
logical :: doDummy = .false.
!    type(BitMaskT), pointer :: flagBitMask(:)
! typedef in ssclib/dss_bitmasks_aux.

!integer(int16), allocatable :: uniqueCombos(:)
!type(TableT) :: eventTab

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '  aaa calcCcdSpatialMap'
    call message(AppMsg, Verbose, '      Calculating the spatial-only &
      &exposure map for this CCD and PI-PATTERN filter.')

if (doDummy) then
  call warning('calcCcdSpatialMap', "This is a dummy routine! It doesn't return the real ccdSpatialMap.")
  ccdSpatialMap = 1.0
return
end if

    ccdSpatialMap = 0.0 ! default

!    numDssComps = size(dssCompNums)

    numDivsSquared = numDivs * numDivs
    realNumDivsSquared = real(numDivsSquared) ! for later use as a denominator.

    xsize = size(ccdSpatialMap, 1)
    ysize = size(ccdSpatialMap, 2)

    numCcdPixels = xsize * ysize
    vectorShape(1) = numCcdPixels
    arrayShape(1) = xsize
    arrayShape(2) = ysize

    write(ccdStr, '(i2.2)') ccdNodeNum
if (testFlag) print *, '  bbb'

    !...................................................................
    ! Construct the coarse pseudo-event list:

!    set = dataSet('tempEvents.ds', TEMP)
    set = dataSet(tempEventListSetName, TEMP)
if (testFlag) print *, '  ccc'
    coarseTab = addTable(set, 'COARSE', numCcdPixels)
if (testFlag) print *, '  ddd0'
    call createCoarsePseudoEvlist(coarseTab, xsize, ysize, timeAttWcs)
if (testFlag) print *, '  ddd1'

    !...................................................................

!allocate(uniqueCombos(size(dssCompNums)))
!call determineUniqueFilterCombos(dssPointer, dssCompNums&
!  , (/'X     ', 'Y     ', 'DETX  ', 'DETY  ', 'RAWX  ', 'RAWY  ', 'FLAG  '/)&
!  , uniqueCombos)

!if (testFlag) print *, 'unique spatial combos:', uniqueCombos

    !...................................................................
    ! Extract a selection expression from each DSS component and filter
    ! the coarse list with it, setting bits of RAW_FLAG or FLAG as
    ! appropriate.
if (testFlag) print *, '  eee0'

!    do index = 1, numDssComps
!      compNum = dssCompNums(index)
!      dssComp = dssComponent(dssPointer, compNum)
!      selExpression = selectExpression(dssComp, 'DETX,DETY,X,Y')
if (present(regionTabName)) then
  selExpression = 'region('//trim(regionTabName)//',X,Y)'
else
  selExpression = ''
end if

      call initFilter(selExpression) 
index = 1
      numRowsSel = doSelectModify(coarseTab, 'FLAG', index-1_int16)
!
!      selExpression = selectExpression(dssComp, 'RAWX,RAWY')
write (selExpression, '(a, i3)') 'RAWY > ', rawYLo
      call initFilter(selExpression)
      numRowsSel = doSelectModify(coarseTab, 'RAW_FLAG', index-1_int16)
!    end do

if (testFlag) print *, '  eee1'
    rawFlagCol = column(coarseTab, 'RAW_FLAG')
    rawFlag => int32Data(rawFlagCol)
if (testFlag) call dumpImageToFits(reshape(rawFlag, arrayShape), 'rawFlag.ds')
if (testFlag) print *, '  eee2'
!    do index = 1, numDssComps
!      compNum = dssCompNums(index)
!      dssComp = dssComponent(dssPointer, compNum)

if (testFlag) print *, '  eee3'
!      call getBitMasks(dssComp, 'FLAG', flagBitMask)
!      ! in ssclib/dss_bitmasks_aux.

if (testFlag) print *, '  fff'
      call calcFlaggedMask(patternNum, badPixelMap, flagBitMask, flaggedCcdMask)
if (testFlag) call dumpImageToFits(flaggedCcdMask, 'flaggedCcdMask.ds')
if (testFlag) print *, '  ggg'

!*** check size of flaggedCcdMask??

!      deallocate(flagBitMask)

index = 1
      where(.not. reshape(flaggedCcdMask, vectorShape))&
        rawFlag = ibclr(rawFlag, index - 1)

      deallocate(flaggedCcdMask)
!    end do
if (testFlag) call dumpImageToFits(reshape(rawFlag, arrayShape), 'rawFlagPlusMask.ds')

    !...................................................................
    ! Identify all 'edge' pixels, ie those neighbours of any edge
    ! created by DETX/Y or X/Y selection. These pixels have the
    ! possibility of taking exposure values intermediate between 0 and 1.
    !
    ! ...but first have to see whether this fine-scale correction is
    ! necessary and required:

    allocate(totalEdgeMap(xsize, ysize))
    totalEdgeMap = .false.

    weightCol = column(coarseTab, 'WEIGHT')
    weight => real32Data(weightCol)
    flagCol = column(coarseTab, 'FLAG')
    flag => int32Data(flagCol)
if (testFlag) call dumpImageToFits(reshape(flag, arrayShape), 'flag.ds')

if (testFlag) print *, '  hhh calcCcdSpatialMap: doing fine-scale correction?'
    if (xyDitherWasDone .and. .not. pleaseCorrectDither) then
      call warning('bug26',"This is not very serious, but nevertheless &
        &should not happen. Please contact the code developer.")
    else if (xyDitherWasDone) then
if (testFlag) print *, '  iii calcCcdSpatialMap: yes, doing fine-scale correction.'
      !.................................................................
      ! Yes, fine-scale correction is both necessary and required.

      allocate(selArray(xsize, ysize), figureEdges(xsize, ysize)&
        , groundEdges(xsize, ysize))

!      do index = 1, numDssComps
index = 1
        selArray = reshape(btest(flag, int(index - 1, int16))&
          , arrayShape)

        call findEdges(selArray, figureEdges, groundEdges)
        ! in ssclib/array_utils.

        totalEdgeMap = totalEdgeMap .or. figureEdges .or. groundEdges
!      end do
      deallocate(selArray, figureEdges, groundEdges)
if (testFlag) print *, '  jjj'

      !.................................................................
      ! If there are >0 'edge' pixels, construct the fine-scale
      ! pseudo-event list:

      numEdgePixels = count(totalEdgeMap)
if (testFlag) print *, '  jjj0 numEdgePixels=', numEdgePixels
      if (numEdgePixels > 0) then
if (testFlag) print *, '  kkk'
        numFineRows = numDivsSquared * numEdgePixels
        fineTab = addTable(set, 'FINE', numFineRows)
        call createFinePseudoEvlist(coarseTab, fineTab, totalEdgeMap&
          , numDivs, timeAttWcs)

        !...............................................................
        ! Filter the fine list on DETX, DETY, X and Y (there's no need
        ! to repeat the RAWX/Y filtration, because the RAWX/Y regions
        ! don't 'cut through' the middle of ccd pixels, hence the micro-
        ! map for each ccd pixel is either all true or all false wrt
        ! the RAWX/Y filter):
 
!        do index = 1, numDssComps
!          compNum = dssCompNums(index)
!          dssComp = dssComponent(dssPointer, compNum)
!          selExpression = selectExpression(dssComp, 'DETX,DETY,X,Y')
!          call initFilter(selExpression) 
!          numRowsSel = doSelectModify(fineTab, 'FINEFLAG'&
!            , int(index - 1, int16))
!        end do
if (testFlag) print *, '  lll'

        !...............................................................
        ! Use the fine list to calculate coarse-list WEIGHT values:
 
        fineFlagCol = column(fineTab, 'FINEFLAG')
        fineFlag => int32Data(fineFlagCol)
        rowInCoarseCol = column(fineTab, 'CRSE_ROW')
        rowInCoarse => int32Data(rowInCoarseCol)

        allocate(finalFlags(numDivsSquared))
        do i = 1, numEdgePixels
          fineRow = (i - 1) * numDivsSquared
          coarseRow = rowInCoarse(fineRow)

          finalFlags = fineFlag(fineRow:fineRow+numDivsSquared-1)
          finalFlags = iand(finalFlags, rawFlag(coarseRow))
          weight(coarseRow) = count(finalFlags > 0) / realNumDivsSquared
        end do
        deallocate(finalFlags)
        call release(fineTab)
      end if ! numEdgePixels > 0.
    end if ! xyDitherWasDone and correctDither required.
if (testFlag) print *, '  mmm'

    !...................................................................
    ! Finally, for all 'edge' pixels, leave the weight values as
    ! calculated; for all the others, the weight should be 0 where neither
    !the RAW or the non-RAW filters passed the pixel for any component,
    !and 1 otherwise.

    where(.not. reshape(totalEdgeMap, vectorShape)&
    .and. (iand(flag, rawFlag) > 0))&
      weight = 1.0

    deallocate(totalEdgeMap)
    ccdSpatialMap = reshape(weight, arrayShape)
    call release(set)

    call message(AppMsg, Noisy, '    Leaving subroutine &
      &calcCcdSpatialMap.')
if (testFlag) print *, '  zzz calcCcdSpatialMap'

  end subroutine calcCcdSpatialMap

  !---------------------------------------------------------------------
  subroutine createCoarsePseudoEvlist(coarseTab, xsize, ysize, timeAttWcs)
  !---------------------------------------------------------------------
  ! NOTE:
  !	(i) It is required that the cal is set up with the appropriate
  ! instrument, ccdChipId, ccdNodeId and ccdModeId beforehand.
!*** check this.

  !	(ii) This routine calls CAL_SetState(randomize=.false.).
  !	(iii) The input argument 'wcs' should be the wcs of the X and
  ! Y columns of the event list.
  !---------------------------------------------------------------------

    implicit none

    type(TableT),       intent(in) :: coarseTab
    integer(int16),     intent(in) :: xsize, ysize
    type(localTAWtype), intent(in) :: timeAttWcs

    integer(int32) :: vectorShape(1), numCcdPixels
!    real(double) :: focalLengthMm
!    real(single) :: mmToDetUnits
real(double) :: mmToDetUnits
    type(ColumnT) ::&
      rawXcol, rawYcol, detXcol, detYcol, xCol, yCol, rawFlagCol, flagCol, weightCol
    integer(int16), pointer :: rawX(:), rawY(:)
integer(int16) :: i
    integer(int16), allocatable :: rawXarray(:,:), rawYarray(:,:)
    real(double), pointer :: cc2X(:), cc2Y(:), cc2Z(:)
    real(single), pointer, dimension(:) ::&
      detX, detY, x, y, weight
    real(double), allocatable :: tanX(:), tanY(:)
    integer(int32), pointer :: flag(:), rawFlag(:)
    logical :: previousRandomizeValue
!    real(double) :: timeStamp
!    type(SpacecraftAttitudeType) :: scAttitude
!    type(WcsT) :: wcs

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '    aaa createCoarsePseudoEvlist'

    vectorShape(1) = xsize * ysize
    numCcdPixels = numberOfRows(coarseTab)

!    call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLengthMm)
!    mmToDetUnits = RadToDeg * (3600.0 / focalLengthMm) / detUnitArcsec
mmToDetUnits = calcMmToDetConversionFactor()
if (testFlag) print *, '    bbb'

    !...................................................................
    ! Add columns RAWX and RAWY. Then set up two temp ccd chip arrays, each of dimensions (xsize, ysize), one containing RAWX values, the other RAWY. Finally, fill the RAWX/Y columns with the appropriate reshaped temp array:

    rawXcol = addColumn(coarseTab, 'RAWX', INTEGER16)
    rawYcol = addColumn(coarseTab, 'RAWY', INTEGER16)
if (testFlag) print *, '    ccc'

    allocate(rawXarray(xsize, ysize))
    rawXarray = spread((/(i,i=1,xsize)/), 2, ysize)
    rawX => int16Data(rawXcol)
    rawX = reshape(rawXarray, vectorShape)
    deallocate(rawXarray)
if (testFlag) print *, '    ddd'

    allocate(rawYarray(xsize, ysize))
    rawYarray = spread((/(i,i=1,ysize)/), 1, xsize)
    rawY => int16Data(rawYcol)
    rawY = reshape(rawYarray, vectorShape)
    deallocate(rawYarray)
if (testFlag) print *, '    eee'

    !...................................................................
    ! Convert RAWX/Y to DETX/Y:

    call CAL_getState(randomize=previousRandomizeValue)
    call CAL_setState(randomize=.false.) ! probably has already been done somewhere in the calling program, but there's no harm in also doing it here, just to make sure. If we had a zillion fake events then randomizing would be a good way to estimate the fraction of real events passed by the filter from a given pixel. But we are only using 1 event per pixel in the coarse pass. It might however provide a good alternative way of doing the second, fine-scale pass...
if (testFlag) print *, '    fff'

    call CAL_rawXY2mm(rawX, rawY, cc2X, cc2Y, cc2Z, coordSysType=CAMCOORD2)
    call release(rawXcol)
    call release(rawYcol)
    detXcol = addColumn(coarseTab, 'DETX', REAL32)
    detYcol = addColumn(coarseTab, 'DETY', REAL32)
    detX => real32Data(detXcol)
    detY => real32Data(detYcol)
    detX = real(cc2X * mmToDetUnits, single)
    detY = real(cc2Y * mmToDetUnits, single)
    call CAL_releaseMemory(cc2X, cc2Y, cc2Z)
if (testFlag) print *, '    ggg'

    !...................................................................
    ! Now convert DETX/Y to X/Y:

    allocate(tanX(numCcdPixels), tanY(numCcdPixels))
!    call instToTan(tanX, tanY, timeAttWcs%wcs%x%refWorld&
!      , timeAttWcs%wcs%y%refWorld, timeAttWcs%attitude&
!      , timeAttWcs%timeStamp, detX, detY)

if (testFlag) then
  print *, '    timeAttWcs%wcs%x%refWorld', timeAttWcs%wcs%x%refWorld
  print *, '    timeAttWcs%wcs%y%refWorld', timeAttWcs%wcs%y%refWorld
  print *, '    timeAttWcs%attitude', timeAttWcs%attitude
  print *, '    timeAttWcs%timeStamp', timeAttWcs%timeStamp
  print *, '    minval(detX)', minval(detX)
  print *, '    maxval(detX)', maxval(detX)
end if

call detToTan(real(detX, double), real(detY, double)&
 , timeAttWcs%wcs%x%refWorld * PiOn180&
 , timeAttWcs%wcs%y%refWorld * PiOn180, timeAttWcs%attitude&
 , timeAttWcs%timeStamp, tanX, tanY)
if (testFlag) print *, '    hhh'

    call release(detXcol)
    call release(detYcol)
    xCol = addColumn(coarseTab, 'X', REAL32)
    yCol = addColumn(coarseTab, 'Y', REAL32)
    x => real32Data(xCol)
    y => real32Data(yCol)
    call tanToPixels(timeAttWcs%wcs, tanX, tanY, x, y) ! in ssclib/wcs_aux
if (testFlag) print *, '    iii'

    deallocate(tanX, tanY)
    call release(xCol)
    call release(yCol)
if (testFlag) print *, '    jjj'

    !...................................................................
    ! Add columns FLAG and WEIGHT:

    flagCol = addColumn(coarseTab, 'FLAG', INTEGER32)
    flag => int32Data(flagCol)
    flag = 0
    call release(flagCol)
    rawFlagCol = addColumn(coarseTab, 'RAW_FLAG', INTEGER32)
    rawFlag => int32Data(rawFlagCol)
    rawFlag = 0
    call release(rawFlagCol)
    weightCol = addColumn(coarseTab, 'WEIGHT', REAL32)
    weight => real32Data(weightCol)
    weight = 0.0
    call release(weightCol)
if (testFlag) print *, '    kkk'

    call CAL_setState(randomize=previousRandomizeValue)
if (testFlag) print *, '    zzz createCoarsePseudoEvlist'

  end subroutine createCoarsePseudoEvlist

  !---------------------------------------------------------------------
  subroutine createFinePseudoEvlist(coarseTab, fineTab, totalEdgeMap&
  , numDivs, timeAttWcs)
  !---------------------------------------------------------------------
  ! NOTE:
  !	(i) It is required that the cal is set up with the appropriate
  ! instrument, ccdChipId, ccdNodeId and ccdModeId beforehand.
  !	(ii) This routine calls CAL_SetState(randomize=.false.).
!** check these 2. (ii) seems in any case to be false.

  !	(iii) The input argument 'wcs' should be the wcs of the X and
  ! Y columns of the event list.
  !---------------------------------------------------------------------

    implicit none

    type(TableT),       intent(in) :: coarseTab, fineTab
    logical(bool),      intent(in) :: totalEdgeMap(:,:)
    integer(int16),     intent(in) :: numDivs
    type(localTAWtype), intent(in) :: timeAttWcs

    type(ColumnT) ::&
      rawXcol, rawYcol, detXcol, detYcol, xCol, yCol,&
      rowInCoarseCol, fineDetXcol, fineDetYcol, fineFlagCol
    integer(int16) :: numDivsSquared, xi, yi, xsize, ysize, arrayShape(2)
    integer(int16) :: indices(numDivs)
    integer(int16), pointer :: rawX(:), rawY(:)
    integer(int32), pointer :: rowInCoarse(:), fineFlag(:)
    integer(int32) ::&
      numCcdPixels,&
      numEdgePixels,&
      numFineRows,&
      fineRow,&
      coarseRow,&
      fineGridVectorShape(1),&
      i
    real(single) :: realNumDivs
    type(Point2dT), allocatable ::&
      detXYarray(:,:), fracVector(:), detXYfineGridVector(:)
    type(Point2dT) :: xDiff, yDiff
    ! typedef in ssclib/geometric_types.

    integer :: foreignStatus
    real(single), pointer, dimension(:) ::&
      detX, detY, x, y, fineDetX, fineDetY
    real(double), allocatable :: tanX(:), tanY(:)
!logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

    xsize = size(totalEdgeMap, 1)
    ysize = size(totalEdgeMap, 2)
    arrayShape(1) = xsize
    arrayShape(2) = ysize
    numCcdPixels = numberOfRows(coarseTab)
    numDivsSquared = numDivs * numDivs
    realNumDivs = real(numDivs) ! for later use as a denominator.
    numEdgePixels = count(totalEdgeMap)
    numFineRows = numDivsSquared * numEdgePixels

    rawXcol = column(coarseTab, 'RAWX')
    rawYcol = column(coarseTab, 'RAWY')
    rawX => int16Data(rawXcol)
    rawY => int16Data(rawYcol)

    rowInCoarseCol = addColumn(fineTab, 'CRSE_ROW', INTEGER32)
    rowInCoarse => int32Data(rowInCoarseCol)
    fineRow = 0
    do coarseRow = 0, numCcdPixels - 1
    if (.not. totalEdgeMap(rawX(coarseRow), rawY(coarseRow))) cycle
      rowInCoarse(fineRow:fineRow+numDivsSquared-1) = coarseRow
      fineRow = fineRow + numDivsSquared
    end do
    call release(rowInCoarseCol)
    call release(rawXcol)
    call release(rawYcol)

    !. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
    ! Convert coarse list DETX/Y into a vector-valued array:

    detXcol = column(coarseTab, 'DETX')
    detYcol = column(coarseTab, 'DETY')
    detX => real32Data(detXcol)
    detY => real32Data(detYcol)
    allocate(detXYarray(xsize, ysize))
    detXYarray%x = reshape(detX, arrayShape)
    detXYarray%y = reshape(detY, arrayShape)
    call release(detXcol)
    call release(detYcol)

    !. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
    ! Now interpolate. The DETX/Y values of the coarse table, ie of the ccd pixels, are presumed to apply to the centre of the pixel. Now we are dividing up each 'coarse' pixel into a numDivs * numDivs array of 'fine' pixels, and we want to interpolate DETX/Y values for these fine pixels.

    fineGridVectorShape(1) = numDivsSquared

    rawXcol = column(coarseTab, 'RAWX')
    rawYcol = column(coarseTab, 'RAWY')
    rawX => int16Data(rawXcol)
    rawY => int16Data(rawYcol)
    allocate(fracVector(numDivsSquared), detXYfineGridVector(numDivsSquared))
    fineDetXcol = addColumn(fineTab, 'DETX', REAL32)
    fineDetYcol = addColumn(fineTab, 'DETY', REAL32)
    fineDetX => real32Data(fineDetXcol)
    fineDetY => real32Data(fineDetYcol)
    rowInCoarseCol = column(fineTab, 'CRSE_ROW')
    rowInCoarse => int32Data(rowInCoarseCol)
    do i = 1, numEdgePixels
      fineRow = (i - 1) * numDivsSquared
      coarseRow = rowInCoarse(fineRow)
      xi = rawX(coarseRow)
      yi = rawY(coarseRow)
      call getGridDiffs(detXYarray, int(xi), int(yi), xDiff, yDiff, foreignStatus)
!*** test foreignStatus

      ! Prepare a list of relative positions for each 'fine' pixel:

      indices = (/(i,i=1,numDivs)/)
      fracVector%x = reshape(&
        spread(&
          fracVal(-0.5, 0.5, (indices - 0.5) / realNumDivs), 2, numDivs&
        ), fineGridVectorShape&
      )
      fracVector%y = reshape(&
        spread(&
          fracVal(-0.5, 0.5, (indices - 0.5) / realNumDivs), 1, numDivs&
        ), fineGridVectorShape&
      )
      ! fracVal() is in ssc_misc.

      ! Now get the list of interpolations:
      call interpolateGridDiff(detXYarray(xi, yi), xDiff, yDiff, fracVector&
        , detXYfineGridVector)

      fineDetX(fineRow:fineRow+numDivsSquared-1) = detXYfineGridVector%x
      fineDetY(fineRow:fineRow+numDivsSquared-1) = detXYfineGridVector%y
    end do

    call release(rawXcol)
    call release(rawYcol)
    call release(rowInCoarseCol)
    deallocate(fracVector, detXYfineGridVector)

    !. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
    ! Now convert DETX/Y to X/Y:

    allocate(tanX(numFineRows), tanY(numFineRows))
!    call instToTan(tanX, tanY, timeAttWcs%wcs%x%refWorld&
!      , timeAttWcs%wcs%y%refWorld, timeAttWcs%attitude&
!      , timeAttWcs%timeStamp, fineDetX, fineDetY)
call detToTan(real(fineDetX, double), real(fineDetY, double)&
 , timeAttWcs%wcs%x%refWorld * PiOn180&
 , timeAttWcs%wcs%y%refWorld * PiOn180, timeAttWcs%attitude&
 , timeAttWcs%timeStamp, tanX, tanY)

    call release(fineDetXcol)
    call release(fineDetYcol)
    xCol = addColumn(fineTab, 'X', REAL32)
    yCol = addColumn(fineTab, 'Y', REAL32)
    x => real32Data(xCol)
    y => real32Data(yCol)
!    call tanToPixels(timeAttWcs%wcs, tanX, tanY, x, y) ! in ssclib/wcs_aux
call l_tanToPixelsVector(timeAttWcs%wcs, tanX, tanY, x, y)

    call release(xCol)
    call release(yCol)

    !. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
    ! Add FLAG column:

    fineFlagCol = addColumn(fineTab, 'FINEFLAG', INTEGER32)
    fineFlag => int32Data(fineFlagCol)
    fineFlag = 0
    call release(fineFlagCol)

  end subroutine createFinePseudoEvlist

  !---------------------------------------------------------------------
  subroutine l_tanToPixelsVector(wcs, tanX, tanY, xPixels, yPixels)
  !---------------------------------------------------------------------
  ! This transform is only valid for CTYPE1='RA---TAN' and CTYPE2='DEC--TAN'.
  !
  ! The tan system coordinates are aligned with the ra/dec world coordinate
  ! system, but are expressed in natural units rather than in whatever units
  ! are specified for the world coordinates. In addition, the tanX axis
  ! points to the right (in the direction of decreasing ra) and the tanY
  ! axis points upwards (in the direction of increasing dec) when the view
  ! is from the centre of the projection sphere outwards.
  !
  ! Numerical examples:
  !
  !	wcs%x%ctype = 'RA---TAN'
  !	wcs%y%ctype = 'DEC--TAN'
  !	wcs%x%refWorld = 25 degrees ra
  !	wcs%y%refWorld = 0 degrees dec
  !
  !	In this case, a position with ra = 30 deg and dec = 0 would give
  !	tanX = -tan(30-25) = -0.0875, tanY = 0; whereas a position with
  !	ra = 0 deg and dec = 7 would give tanX = 0, tanY = tan(7) = 0.123.
  !
  ! As described in Greisen & Calabretta A&A 395, 1061 (2002) equation (3),
  ! the relation between pixel coordinates p_i and intermediate world
  ! coordinates x_i is
  !
  !	      _ N
  !	x_i = >_  CDi_j (p_j - CRPIXj).
  !	     j = 1
  !
  ! This is inverted in the present subroutine, after x_i is first converted
  ! from my tan system.
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: tanX(:), tanY(size(tanX))
    type(WcsT),   intent(in)  :: wcs
    real(single), intent(out) :: xPixels(size(tanX)), yPixels(size(tanX))

real(double), allocatable ::&
intermediateWorldX(:), intermediateWorldY(:)
    real(double) ::&
      intermediateWorldXToTanX, intermediateWorldYToTanY,&
!      intermediateWorldX(size(tanX)), intermediateWorldY(size(tanX)),&
      det, invMatrix(2,2)
!    real(single) :: relPixelsX(size(tanX)), relPixelsY(size(tanX))
real(single), allocatable :: relPixelsX(:), relPixelsY(:)

    logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '      aaa l_tanToPixelsVector'

allocate(intermediateWorldX(size(tanX)), intermediateWorldY(size(tanX)))
allocate(relPixelsX(size(tanX)), relPixelsY(size(tanX)))

!!!    call checkWcsType(wcs)
if (testFlag) print *, '      bbb'

    select case(wcs%x%worldUnit)
    case('deg')
      intermediateWorldXToTanX = -PiOn180 ! note the negation
    case default
call error('notSupported',"")
    end select
if (testFlag) print *, '      ccc'

    select case(wcs%y%worldUnit)
    case('deg')
      intermediateWorldYToTanY = PiOn180
    case default
call error('notSupported',"")
    end select
if (testFlag) print *, '      ddd'

    intermediateWorldX = tanX / intermediateWorldXToTanX
    intermediateWorldY = tanY / intermediateWorldYToTanY
if (testFlag) print *, '      eee'

    det = wcs%x%matrixOnAxis  * wcs%y%matrixOnAxis&
        - wcs%x%matrixOffAxis * wcs%y%matrixOffAxis
if (det == 0.0) call error('singularWcsRotMatrix',"")
if (testFlag) print *, '      fff'

    invMatrix(1,1) =  wcs%y%matrixOnAxis  / det		! 1st index row,
    invMatrix(1,2) = -wcs%y%matrixOffAxis / det		! 2nd index col.
    invMatrix(2,1) = -wcs%x%matrixOffAxis / det
    invMatrix(2,2) =  wcs%x%matrixOnAxis  / det
if (testFlag) print *, '      ggg'

    relPixelsX = invMatrix(1,1) * intermediateWorldX&
               + invMatrix(1,2) * intermediateWorldY
    relPixelsY = invMatrix(2,1) * intermediateWorldX&
               + invMatrix(2,2) * intermediateWorldY
if (testFlag) print *, '      hhh'

    xPixels = relPixelsX + wcs%x%refPixels
    yPixels = relPixelsY + wcs%y%refPixels
if (testFlag) print *, '      zzz l_tanToPixelsVector'

  end subroutine l_tanToPixelsVector

  !---------------------------------------------------------------------
  subroutine calcFlaggedMask(patternNum, badPixelMap, flagBitMask&
  , flaggedCcdMask)
  !---------------------------------------------------------------------

!*** Don't forget to deallocate flaggedCcdMask after use.

!*** say what cal calls beneath, and what cal variables need to be set.

    implicit none

    integer(int8),  intent(in) :: patternNum
    integer(int8),  pointer    :: badPixelMap(:,:)
    type(BitMaskT), intent(in) :: flagBitMask(:)
    ! typedef in ssclib/dss_bitmasks_aux.
    logical(bool),  pointer    :: flaggedCcdMask(:,:)

    integer(int16) :: xSize, ySize
    integer :: numBits, numMasks, maskNum, bitNum
    logical(bool) :: noFilter
    integer(int32), allocatable :: flaggedMap(:,:)
    logical(bool), allocatable :: badMask(:,:), tempFlagMask(:,:)
    integer(int32) :: instrumentId

logical :: testFlag = .false.
!character(1) :: dummyStr

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '    aaa calcFlaggedMask'
    call message(AppMsg, Noisy, '      Entering subroutine calcFlaggedMask.')

    numBits = bit_size(flagBitMask%onBits) ! f90 intrinsic fn
    numMasks = size(flagBitMask)
if (testFlag) print *, '    bbb'
if (testFlag) print *, '    bbb0 numMasks=', numMasks
if (testFlag) print *, '    bbb1 flagBitMask%onBits=', flagBitMask%onBits
if (testFlag) print *, '    bbb2 flagBitMask%offBits=', flagBitMask%offBits
    if (numMasks <= 0) then
      noFilter = .true.
!call error('waitACottonPickinMinute',"")
    else
      if (all(flagBitMask%onBits == 0) .and. all(flagBitMask%offBits == 0)) then
        noFilter = .true.
      else
        noFilter = .false.
      end if
    end if
if (testFlag) print *, '    ccc'

    call CAL_getState(instrument=instrumentId)
if (testFlag) print *, '    ddd'

    if (instrumentId == EPN) then
      xSize = EPN_MAX_X
      ySize = EPN_MAX_Y
      allocate(flaggedCcdMask(xSize,ySize))
      allocate(flaggedMap( xSize,ySize))
      allocate(badMask(    xSize,ySize))
      call calcPnFlagMap(patternNum, badPixelMap, badMask, flaggedMap)

!write(dummyStr, '(i1)') patternNum
!call dumpImageToFits(flaggedMap, 'flags_pattern'//dummyStr//'_ccd05.ds')

    else if (instrumentId == EMOS1 .or. instrumentId == EMOS2) then
      xSize = EMOS_MAX_X
      ySize = EMOS_MAX_Y
      allocate(flaggedCcdMask(xSize,ySize))
      allocate(flaggedMap( xSize,ySize))
      allocate(badMask(    xSize,ySize))
      call calcMosFlagMap(noFilter, badPixelMap, badMask, flaggedMap)

    else
call error('badInstrument',"The instrument was not recognized.")
    end if
if (testFlag) print *, '    eee'

    flaggedCcdMask = badMask
!  if (noFilter) return ! because there was no filtering on FLAG.
if (noFilter) then
  if (testFlag) print *, '    zzz calcFlaggedMask'
  return ! because there was no filtering on FLAG.
end if

if (testFlag) print *, '    fff'
    !...................................................................
    ! Finally, do the filtering.
    !
    ! The FLAG filtering expression contained in the vector flagFilter
    ! is interpreted as follows. Suppose we set boolean flags onBitsMatch_i,
    ! offBitsMatch_i and allBitsMatch_i for pixel (xi,yi) as follows:
    !
    !    onBitsMatch_i&
    !      = (iand(flagFilter(i)%onBits, flaggedMap(xi,yi))&
    !         == flagFilter(i)%onBits)
    !
    !    offBitsMatch_i&
    !      = (iand(flagFilter(i)%offBits, not(flaggedMap(xi,yi)))&
    !         == flagFilter(i)%offBits)
    !
    !    allBitsMatch_i = onBitsMatch_i .and. offBitsMatch_i.
    !
    ! The value of this pixel on the mask depends on the result of
    ! the following boolean expression:
    !
    ! flaggedCcdMask = (allBitsMatch_1 || allBitsMatch_2...).

    if (bitMaskFilterIsBad(flagBitMask))& ! in ssclib/dss_bitmasks_aux.
call error('badFlagFilter',"The bit-mask filter on column FLAG which was &
  &just obtained from the data subspace was not well-formed.")

if (testFlag) print *, '    ggg'

    flaggedCcdMask = .false.
    allocate(tempFlagMask(xSize,ySize))
if (testFlag) print *, '    hhh'
    do maskNum = 1, numMasks
      tempFlagMask = .true.
      do bitNum = 0, numBits - 1
        if (btest(flagBitMask(maskNum)%onBits, bitNum)) then ! the same bit
        ! is required to be set in flaggedMap. Set tempFlagMask to .false.
        ! if it is not:

          where(.not. btest(flaggedMap, bitNum)) tempFlagMask = .false.

        else if (btest(flagBitMask(maskNum)%offBits, bitNum)) then ! the same
        ! bit is required to be unset in flaggedMap. Set tempFlagMask to
        ! .false. if it is not:

          where(btest(flaggedMap, bitNum)) tempFlagMask = .false.
        end if
      end do
      ! tempFlagMask now holds a record of those pixels for which all the
      ! flagFilter%onBits were set and all the flagFilter%offBits were unset.

      flaggedCcdMask = flaggedCcdMask .or. tempFlagMask
    end do
if (testFlag) print *, '    iii'

    ! Finally:
    flaggedCcdMask = flaggedCcdMask .and. badMask

    call message(AppMsg, Noisy, '      Leaving subroutine calcFlaggedMask.')
if (testFlag) print *, '    zzz calcFlaggedMask'

  end subroutine calcFlaggedMask

  !---------------------------------------------------------------------
  subroutine calcPnFlagMap(patternNum, badPixelMap, badMask, flagMap)
  !---------------------------------------------------------------------
  ! This subroutine returns two CCD arrays, badMask and flagMap. The pixels
  ! of badMask are set TRUE if it is possible to obtain a cosmic x-ray event
  ! on that pixel, otherwise FALSE. Hence bad pixels are set FALSE, as are
  ! all pixels outside the window area. At this time I am undecided how to
  ! handle out-of-FOV pixels: I have chosen not to set them to FALSE in
  ! badMask at this time but rather to flag them in flagMap.
  !
  ! The pixels of the second array, flagMap, are 4-byte-integer valued, the
  ! same as the FLAG column of the event list. There is a list of these FLAG
  ! bits which are purely spatial, viz:
  !
  !	-----------------------------------------------
  !	Bit num		Name:
  !	(from 0):
  !	-----------------------------------------------
  !	   2		CLOSE_TO_CCD_WINDOW
  !	   3		ON_OFFSET_COLUMN
  !	   4		NEXT_TO_OFFSET_COLUMN
  !	   5		CLOSE_TO_ONBOARD_BADPIX
  !	   6		CLOSE_TO_BRIGHTPIX
  !	   8		CLOSE_TO_DEADPIX
  !	  16		OUT_OF_FOV
  !	-----------------------------------------------
  !
  ! The array flagMap has the corresponding bits of the appropriate pixels
  ! set.
  !---------------------------------------------------------------------

!*** say what cal calls beneath, and what cal variables need to be set.
    implicit none

    integer(int8),  intent(in)  :: patternNum
    integer(int8),  pointer     :: badPixelMap(:,:)
    logical(bool),  intent(out) :: badMask(:,:)
    integer(int32), intent(out) :: flagMap(size(badMask,1),size(badMask,2))

    integer(int16) ::&
      xSize, ySize
    integer(int32) :: evatts(3), myBits(3), bit, i, evattFlag

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '      aaa calcPnFlagMap'
    call message(AppMsg, Noisy, '      Entering subroutine calcPnFlagMap.')

    xSize = EPN_MAX_X
    ySize = EPN_MAX_Y
    if (size(badMask,1) .ne. xSize .or. size(badMask,2) .ne. ySize)&
call error('bug24',"Please contact the code developer.")

!*** check on size of badPixelMap

    badMask = .false. ! default
    flagMap = 0    ! default

if (testFlag) print *, '      bbb'

    !...................................................................
    ! Get window borders:

    where(.not. btest(badPixelMap(1:xSize,1:ySize), OUTSIDE_WINDOW_BIT))&
      badMask = .true.

    !...................................................................
    ! Flag or delete OOFOV pixels:

!print *, 'EVATT_OUT_OF_FOV=', EVATT_OUT_OF_FOV
    where(btest(badPixelMap(1:xSize,1:ySize), OUTSIDE_FOV_BIT))
!!!      badMask = .false.
      flagMap = ior(flagMap, EVATT_OUT_OF_FOV)
    end where
if (testFlag) print *, '      ccc'

    !...................................................................
    ! Process the window borders further:

    select case(patternNum)
    case(0)
    case(1)
      where(btest(iand(badPixelMap(1:xSize,2:ySize+1)&
        , not(badPixelMap(1:xSize,1:ySize))), OUTSIDE_WINDOW_BIT))&
        badMask = .false.
    case(2)
      where(btest(iand(badPixelMap(2:xSize+1,1:ySize)&
        , not(badPixelMap(1:xSize,1:ySize))), OUTSIDE_WINDOW_BIT))&
        badMask = .false.
    case(3)
      where(btest(iand(badPixelMap(1:xSize,0:ySize-1)&
        , not(badPixelMap(1:xSize,1:ySize))), OUTSIDE_WINDOW_BIT))&
        badMask = .false.
    case(4)
      where(btest(iand(badPixelMap(0:xSize-1,1:ySize)&
        , not(badPixelMap(1:xSize,1:ySize))), OUTSIDE_WINDOW_BIT))&
        badMask = .false.
    case default
    end select
if (testFlag) print *, '      ddd'

!print *, 'evattIsSet(CLOSE_TO_CCD_WINDOW_I)?', evattIsSet(CLOSE_TO_CCD_WINDOW_I)
!print *, 'patternNum=', patternNum
!print *, 'EVATT_CLOSE_TO_CCD_WINDOW=', EVATT_CLOSE_TO_CCD_WINDOW

    if (evattIsSet(CLOSE_TO_CCD_WINDOW_I)) then
      if (patternNum < 5) then
!*** epevents doesn't at present seem to set this flag for any events but singles and doubles.
        if (patternNum .ne. 1) then
          where(btest(iand(badPixelMap(1:xSize,2:ySize+1)&
            , not(badPixelMap(1:xSize,1:ySize))), OUTSIDE_WINDOW_BIT))&
            flagMap = ior(flagMap, EVATT_CLOSE_TO_CCD_WINDOW)
        end if
        if (patternNum .ne. 2) then
          where(btest(iand(badPixelMap(2:xSize+1,1:ySize)&
            , not(badPixelMap(1:xSize,1:ySize))), OUTSIDE_WINDOW_BIT))&
            flagMap = ior(flagMap, EVATT_CLOSE_TO_CCD_WINDOW)
        end if
        if (patternNum .ne. 3) then
          where(btest(iand(badPixelMap(1:xSize,0:ySize-1)&
            , not(badPixelMap(1:xSize,1:ySize))), OUTSIDE_WINDOW_BIT))&
            flagMap = ior(flagMap, EVATT_CLOSE_TO_CCD_WINDOW)
        end if
        if (patternNum .ne. 4) then
          where(btest(iand(badPixelMap(0:xSize-1,1:ySize)&
            , not(badPixelMap(1:xSize,1:ySize))), OUTSIDE_WINDOW_BIT))&
            flagMap = ior(flagMap, EVATT_CLOSE_TO_CCD_WINDOW)
        end if
      end if
    end if
if (testFlag) print *, '      eee'

    !...................................................................
    ! Now the bad pixels and their borders:

    myBits = (/BRIGHTPIX_BIT, ONBOARD_BADPIX_BIT, DEADPIX_BIT/)
    evAtts = (/EVATT_CLOSE_TO_BRIGHTPIX, EVATT_CLOSE_TO_ONBOARD_BADPIX&
      , EVATT_CLOSE_TO_DEADPIX/)

    do i = 1, size(myBits)
      bit = myBits(i)
      evattFlag = evAtts(i)
!print *, 'evAtts(',i,')=', evAtts(i)

      where(btest(badPixelMap(1:xSize,1:ySize), bit)) badMask = .false.

    if (patternNum .ne. 0) cycle
!*** epevents seems at present to flag badpix neighbours only for singles. 

      where(btest(badPixelMap(1:xSize,  2:ySize+1), bit)&
       .or. btest(badPixelMap(2:xSize+1,1:ySize  ), bit)&
       .or. btest(badPixelMap(1:xSize,  0:ySize-1), bit)&
       .or. btest(badPixelMap(0:xSize-1,1:ySize  ), bit))&
         flagMap = ior(flagMap, evattFlag)
    end do

!*** under the present epevents scheme:
! EVATT_CLOSE_TO_DEADPIX trumps EVATT_CLOSE_TO_BRIGHTPIX trumps EVATT_CLOSE_TO_ONBOARD_BADPIX

!** trumps EVATT_CLOSE_TO_CCD_WINDOW:

    if (patternNum == 0) then
!print *, 'EVATT_CLOSE_TO_DEADPIX', EVATT_CLOSE_TO_DEADPIX
!print *, 'EVATT_CLOSE_TO_BRIGHTPIX', EVATT_CLOSE_TO_BRIGHTPIX
!print *, 'EVATT_CLOSE_TO_ONBOARD_BADPIX', EVATT_CLOSE_TO_ONBOARD_BADPIX
!print *, 'EVATT_CLOSE_TO_CCD_WINDOW', EVATT_CLOSE_TO_CCD_WINDOW
    !*** epevents seems at present to flag badpix neighbours only for singles. 
      where(iand(flagMap, EVATT_CLOSE_TO_DEADPIX) .ne. 0)
        flagMap = iand(flagMap, not(EVATT_CLOSE_TO_BRIGHTPIX))
        flagMap = iand(flagMap, not(EVATT_CLOSE_TO_ONBOARD_BADPIX))
!!!        flagMap = iand(flagMap, not(EVATT_CLOSE_TO_CCD_WINDOW))
      end where

      where(iand(flagMap, EVATT_CLOSE_TO_BRIGHTPIX) .ne. 0)
        flagMap = iand(flagMap, not(EVATT_CLOSE_TO_ONBOARD_BADPIX))
!!!        flagMap = iand(flagMap, not(EVATT_CLOSE_TO_CCD_WINDOW))
      end where

!!!      where(iand(flagMap, EVATT_CLOSE_TO_ONBOARD_BADPIX) .ne. 0)
!!!        flagMap = iand(flagMap, not(EVATT_CLOSE_TO_CCD_WINDOW))
!!!      end where
    end if

    !...................................................................
    ! Add bad offset columns and rows:

!print *, 'EVATT_ON_OFFSET_COLUMN', EVATT_ON_OFFSET_COLUMN
!print *, 'EVATT_NEXT_TO_OFFSET_COLUMN', EVATT_NEXT_TO_OFFSET_COLUMN
    where(btest(badPixelMap(1:xSize,1:ySize), BAD_OFFSET_BIT))&
      flagMap = ior(flagMap, EVATT_ON_OFFSET_COLUMN)

    where(btest(badPixelMap(0:xSize-1,1:ySize), BAD_OFFSET_BIT)&
     .or. btest(badPixelMap(2:xSize+1,1:ySize), BAD_OFFSET_BIT))&
      flagMap = ior(flagMap, EVATT_NEXT_TO_OFFSET_COLUMN)


    call message(AppMsg, Noisy, '      Leaving subroutine calcPnFlagMap.')
if (testFlag) print *, '      zzz calcPnFlagMap'

  end subroutine calcPnFlagMap

  !---------------------------------------------------------------------
  subroutine calcMosFlagMap(noFilter, badPixelMap, badMask, flagMap)
  !---------------------------------------------------------------------
  ! This subroutine returns two CCD arrays, badMask and flagMap. The pixels
  ! of badMask are set TRUE if it is possible to obtain a cosmic x-ray event
  ! on that pixel, otherwise FALSE. Hence bad pixels, offset rows and offset
  ! columns are all set FALSE, as are all pixels outside the window area. At
  ! this time I am undecided how to handle out-of-FOV pixels: I have chosen
  ! not to set them to FALSE in badMask at this time but rather to flag them
  ! in flagMap.
  !
  ! The pixels of the second array, flagMap, are 4-byte-integer valued, the
  ! same as the FLAG column of the event list. There is a list of these FLAG
  ! bits which are purely spatial, viz:
  !
  !	-----------------------------------------------
  !	Bit num		Name:
  !	(from 0):
  !	-----------------------------------------------
  !	   1		CLOSE_TO_CCD_BORDER
  !	   2		CLOSE_TO_CCD_WINDOW
  !	   3		CLOSE_TO_NODE_BOUNDARY
  !	   5		CLOSE_TO_ONBOARD_BADPIX
  !	   6		CLOSE_TO_BRIGHTPIX
  !	   8		CLOSE_TO_DEADPIX
  !	   9		CLOSE_TO_BADCOL
  !	  10		CLOSE_TO_BADROW
  !	  16		OUT_OF_FOV
  !	-----------------------------------------------
  !
  ! The array flagMap has the corresponding bits of the appropriate pixels
  ! set.
  !---------------------------------------------------------------------

!*** say what cal calls beneath, and what cal variables need to be set.
    implicit none

    logical(bool),  intent(in)  :: noFilter
    integer(int8),  pointer     :: badPixelMap(:,:)
    logical(bool),  intent(out) :: badMask(:,:)
    integer(int32), intent(out) :: flagMap(size(badMask,1),size(badMask,2))

    integer(int16) ::&
      xSize, ySize
    integer(int32) :: evatts(3), myBits(3), bit, i, evattFlag

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '      aaa calcMosFlagMap'
    call message(AppMsg, Noisy, '      Entering subroutine calcMosFlagMap.')

    xSize = EMOS_MAX_X
    ySize = EMOS_MAX_Y
    if (size(badMask,1) .ne. xSize .or. size(badMask,2) .ne. ySize)&
call error('bug25',"Please contact the code developer.")

!*** check on size of badPixelMap

    badMask = .false. ! default
    flagMap = 0    ! default

if (testFlag) print *, '      bbb'

    !...................................................................
    ! Get window borders:

    where(.not. btest(badPixelMap(1:xSize,1:ySize), OUTSIDE_WINDOW_BIT))&
      badMask = .true.

    !...................................................................
    ! Flag or delete OOFOV pixels:

    where(btest(badPixelMap(1:xSize,1:ySize), OUTSIDE_FOV_BIT))
!!!      badMask = .false.
      flagMap = ior(flagMap, EVATT_OUT_OF_FOV)
    end where
    ! Note also that there is no point setting this attribute if you also
    ! set badMask to .false. at these oofov pixels. You should do one or
    ! the other, not both. Leaving them flagged in flagMap but not .false.
    ! in badMask can be useful if you want to analyse non-x-ray events
    ! outside the FOV.
if (testFlag) print *, '      ccc'

    !...................................................................
    ! Now the bad pixels and their borders:

    myBits = (/BRIGHTPIX_BIT, ONBOARD_BADPIX_BIT, DEADPIX_BIT/)
    evAtts = (/EVATT_CLOSE_TO_BRIGHTPIX, EVATT_CLOSE_TO_ONBOARD_BADPIX&
      , EVATT_CLOSE_TO_DEADPIX/)

    do i = 1, size(myBits)
      bit = myBits(i)
      evattFlag = evAtts(i)

      where(btest(badPixelMap(1:xSize,1:ySize), bit)) badMask = .false.

      where(btest(badPixelMap(1:xSize,  2:ySize+1), bit)&
       .or. btest(badPixelMap(2:xSize+1,1:ySize  ), bit)&
       .or. btest(badPixelMap(1:xSize,  0:ySize-1), bit)&
       .or. btest(badPixelMap(0:xSize-1,1:ySize  ), bit))&
         flagMap = ior(flagMap, evattFlag)
    end do
if (testFlag) print *, '      ddd'

    !...................................................................
    ! Add bad offset columns and rows:

    ! For MOS, one can't get any x-ray events on the offset columns, so just
    ! include them in the mask:
    where(btest(badPixelMap(1:xSize,1:ySize), BAD_OFFSET_BIT))&
      badMask = .false.

    where(btest(badPixelMap(0:xSize-1,1:ySize), BAD_OFFSET_BIT)&
     .or. btest(badPixelMap(2:xSize+1,1:ySize), BAD_OFFSET_BIT))&
      flagMap = ior(flagMap, EVATT_CLOSE_TO_BADCOL)

    where(btest(badPixelMap(1:xSize,0:ySize-1), BAD_OFFSET_BIT)&
     .or. btest(badPixelMap(1:xSize,2:ySize+1), BAD_OFFSET_BIT))&
      flagMap = ior(flagMap, EVATT_CLOSE_TO_BADROW)

    if (noFilter) then
  return ! because there was no filtering on FLAG.
    end if
if (testFlag) print *, '      eee'

    !...................................................................
    ! Process the ccd and window borders:

    if (evattIsSet(CLOSE_TO_CCD_BORDER_I)) then
      flagMap(1,    :) = ior(flagMap(1,    :), EVATT_CLOSE_TO_CCD_BORDER)
      flagMap(xsize,:) = ior(flagMap(xsize,:), EVATT_CLOSE_TO_CCD_BORDER)
      flagMap(:,    1) = ior(flagMap(:,    1), EVATT_CLOSE_TO_CCD_BORDER)
      flagMap(:,ysize) = ior(flagMap(:,ysize), EVATT_CLOSE_TO_CCD_BORDER)
    end if
if (testFlag) print *, '      fff'

    if (evattIsSet(CLOSE_TO_NODE_BOUNDARY_I)) then
      if (.not. any(btest(badPixelMap(1:xSize,1:ySize), NODE_BOUNDARY_BIT)))&
      then

        ! Note that this flag is not used by pn because there is only 1
        ! node there.

        call warning('inconsistentNodeInfo',"The CLOSE_TO_NODE_BOUNDARY &
          &event attribute was set, but lccorr can't find any bits of &
          &the flag mask with this value.")
      else
        where(btest(badPixelMap(1:xSize,1:ySize), NODE_BOUNDARY_BIT))&
          flagMap = ior(flagMap, EVATT_CLOSE_TO_NODE_BOUNDARY)
      end if
    end if
if (testFlag) print *, '      ggg'

    if (evattIsSet(CLOSE_TO_CCD_WINDOW_I)) then
      if (.not. (any(btest(badPixelMap(1,1:ySize), OUTSIDE_WINDOW_BIT))&
      .or. any(btest(badPixelMap(1:xSize,1), OUTSIDE_WINDOW_BIT)))) then

        call warning('pointlessWindowSelection',"The CLOSE_TO_CCD_WINDOW &
          &event attribute is set but no events are flagged with this &
          &value when the window occupies the whole ccd.")

      else
        where(btest(badPixelMap(0:xSize-1,1:ySize), OUTSIDE_WINDOW_BIT)&
         .or. btest(badPixelMap(1:xSize,0:ySize-1), OUTSIDE_WINDOW_BIT))&
          flagMap = ior(flagMap, EVATT_CLOSE_TO_CCD_WINDOW)
        ! Note that for MOS only events at the left and bottom of the
        ! window are so flagged.
      end if
    end if

if (testFlag) print *, '      hhh'

    call message(AppMsg, Noisy, '      Leaving subroutine calcMosFlagMap.')
if (testFlag) print *, '      zzz calcMosFlagMap'

  end subroutine calcMosFlagMap

  !---------------------------------------------------------------------
  subroutine readEvatts(eventSet)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in)  :: eventSet

    type(TableT) :: eventTab

    !-------------------------------------------------------------------
    ! Start

    eventTab = table(eventSet, 'EVENTS')

    evattIsSet = .false._bool ! default

    if (eventAttributeIsSet(eventTab, EVATT_OUT_OF_FOV))&
      evattIsSet(OUT_OF_FOV_I) = .true._bool

    if (eventAttributeIsSet(eventTab, EVATT_CLOSE_TO_CCD_BORDER))&
      evattIsSet(CLOSE_TO_CCD_BORDER_I) = .true._bool

    if (eventAttributeIsSet(eventTab, EVATT_CLOSE_TO_NODE_BOUNDARY))&
      evattIsSet(CLOSE_TO_NODE_BOUNDARY_I) = .true._bool

    if (eventAttributeIsSet(eventTab, EVATT_CLOSE_TO_CCD_WINDOW))&
      evattIsSet(CLOSE_TO_CCD_WINDOW_I) = .true._bool

    call release(eventTab)

  end subroutine readEvatts

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module exposure_map_pcms
